from typing import Optional, Union

import abjad

from ._HarmonicParent import _HarmonicParent


class HarmonicNote(abjad.Note, _HarmonicParent):
    r"""Creates a note with tweaked note head for harmonics. This is a child
    class of |abjad.Note|.

    Basic usage:
        Usage is similar to |abjad.Note|:

        >>> harm = auxjad.HarmonicNote(r"c''4")
        >>> harm.style
        "#'harmonic"
        >>> abjad.show(harm)

        ..  docs::

            \tweak style #'harmonic
            c''4

        ..  figure:: ../_images/HarmonicNote-jslykzpz7en.png

        And similarly to |abjad.Note|, pitch and duration can be input in
        many different ways:

        >>> harm1 = auxjad.HarmonicNote(r"c''4")
        >>> harm2 = auxjad.HarmonicNote(r"c''", 1 / 4)
        >>> harm3 = auxjad.HarmonicNote(12, 0.25)
        >>> harm4 = auxjad.HarmonicNote(12, abjad.Duration(1, 4))
        >>> staff = abjad.Staff([harm1, harm2, harm3, harm4])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \tweak style #'harmonic
                c''4
                \tweak style #'harmonic
                c''4
                \tweak style #'harmonic
                c''4
                \tweak style #'harmonic
                c''4
            }

        ..  figure:: ../_images/HarmonicNote-kdx8pmkdspn.png

    :attr:`style`:
        When instantiating this class, use the keyword argument :attr:`style`
        to set a different type of note head, such as ``"#'harmonic-mixed"``:

        >>> harm = auxjad.HarmonicNote(r"c''4",
        ...                            style="#'harmonic-mixed",
        ...                            )
        >>> harm.style
        "#'harmonic-mixed"
        >>> abjad.show(harm)

        ..  docs::

            \tweak style #'harmonic-mixed
            c''4

        ..  figure:: ../_images/HarmonicNote-z48i5il6vf.png

    :attr:`~abjad.core.Note.Note.multiplier`:
        Similarly to |abjad.Note|, this class can take multipliers:

        >>> harm = auxjad.HarmonicNote(r"c''4",
        ...                            multiplier=(2, 3),
        ...                            )
        >>> harm.multiplier
        abjad.Multiplier(2, 3)
        >>> abjad.show(harm)

        ..  docs::

            \tweak style #'harmonic
            c''4 * 2/3

        ..  figure:: ../_images/HarmonicNote-4qbhly2hfi.png

    Properties:
        All properties of |abjad.Note| are also available to be read. This
        class also includes a new property named :attr:`style`:

        >>> harm = auxjad.HarmonicNote(r"c''4")
        >>> harm.written_pitch
        "c''"
        >>> harm.written_duration
        1/4
        >>> harm.style
        "#'harmonic"

        All these properties can be set to different values after
        initialisation:

        >>> harm.written_pitch = 18
        >>> harm.written_duration = abjad.Duration(1, 8)
        >>> harm.style = "#'harmonic-mixed"
        >>> harm.written_pitch
        "fs''"
        >>> harm.written_duration
        1/8
        >>> harm.style
        "#'harmonic-mixed"

    Setting :attr:`style` to ``'flageolet'``:
        To create a harmonic note with a regular note head and with a flageolet
        circle above it, use the style ``'flageolet'``:

        >>> harm = auxjad.HarmonicNote(r"c''1",
        ...                            style='flageolet',
        ...                            )
        >>> harm.style
        'flageolet'
        >>> abjad.show(harm)

        ..  docs::

            c''1
            \flageolet

        ..  figure:: ../_images/HarmonicNote-4q2q7rz65lj.png

    :attr:`markup`:
        To add a markup expression to the harmonic note, use the :attr:`markup`
        optional keyword argument, which takes strings. By default, the markup
        position is above the harmonic note, but this can be overridden using
        the keyword :attr:`direction`, which can take strings as well as
        ``abjad.Up`` and ``abjad.Down``:

        >>> harm1 = auxjad.HarmonicNote(r"d''1")
        >>> harm2 = auxjad.HarmonicNote(r"d''1",
        ...                             markup='III.',
        ...                             )
        >>> harm3 = auxjad.HarmonicNote(r"d''1",
        ...                             markup='III.',
        ...                             direction=abjad.Down)
        >>> staff = abjad.Staff([harm1, harm2, harm3])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \tweak style #'harmonic
                d''1
                \once \override TextScript.parent-alignment-X = 0
                \once \override TextScript.self-alignment-X = 0
                \tweak style #'harmonic
                d''1
                ^ \markup { III. }
                \once \override TextScript.parent-alignment-X = 0
                \once \override TextScript.self-alignment-X = 0
                \tweak style #'harmonic
                d''1
                _ \markup { III. }
            }

        ..  figure:: ../_images/HarmonicNote-v9uer4i864.png

        Setting :attr:`markup` to ``None`` will remove the markup from the
        note.

        >>> harm = auxjad.HarmonicNote(r"d''1",
        ...                            markup='III.',
        ...                            )
        >>> harm.markup = None
        >>> abjad.show(harm)

        ..  docs::

            \tweak style #'harmonic
            d''1

        ..  figure:: ../_images/HarmonicNote-2gqky0o8dgt.png

    :attr:`centre_markup`:
        When a markup expression is added to the harmonic note by using the
        :attr:`markup` optional keyword argument, it will be automatically
        centred above the note (as the main purpose of this markup is to show
        string numbers). To disable this behaviour, set :attr:`centre_markup`
        to ``False``. Compare:

        >>> harm1 = auxjad.HarmonicNote(r"d''1",
        ...                            markup='III.',
        ...                            )
        >>> abjad.show(harm1)

        ..  docs::

            \once \override TextScript.parent-alignment-X = 0
            \once \override TextScript.self-alignment-X = 0
            \tweak style #'harmonic
            d''1
            ^ \markup { III. }

        ..  figure:: ../_images/HarmonicNote-Vb1lf8wt7O.png

        >>> harm2 = auxjad.HarmonicNote(r"d''1",
        ...                            markup='III.',
        ...                            centre_markup=False,
        ...                            )
        >>> abjad.show(harm2)

        ..  docs::

            \tweak style #'harmonic
            d''1
            ^ \markup { III. }

        ..  figure:: ../_images/HarmonicNote-ZxHdPOas1z.png

    ..  error::

        If another markup is attached to the harmonic note, trying to set the
        :attr:`markup` to ``None`` will raise an :exc:`Exception`:

        >>> harm = auxjad.HarmonicNote(r"d''1")
        >>> abjad.attach(abjad.Markup('test'), harm)
        >>> harm.markup = 'III.'
        >>> harm.markup = None
        Exception: multiple indicators attached to client.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_style", "_direction", "_markup", "_centre_markup")

    ### INITIALISER ###

    def __init__(
        self,
        *arguments,
        multiplier: Optional[abjad.typings.DurationTyping] = None,
        tag: Optional[abjad.Tag] = None,
        style: str = "#'harmonic",
        markup: Optional[str] = None,
        centre_markup: bool = True,
        direction: Union[str, abjad.enums.VerticalAlignment] = "up",
    ) -> None:
        super().__init__(*arguments, multiplier=multiplier, tag=tag)
        self.style = style
        self._direction = direction
        self.centre_markup = centre_markup
        self.markup = markup

    ### PRIVATE METHODS ###

    def _attach_centre_markup(self) -> None:
        r"""Attaches the centre markup tweaks."""
        literal1 = abjad.LilyPondLiteral(r"\once \override TextScript.parent-alignment-X = 0")
        literal2 = abjad.LilyPondLiteral(r"\once \override TextScript.self-alignment-X = 0")
        abjad.attach(literal1, self)
        abjad.attach(literal2, self)

    def _detach_centre_markup(self) -> None:
        r"""Detaches the centre markup tweaks."""
        literal1 = abjad.LilyPondLiteral(r"\once \override TextScript.parent-alignment-X = 0")
        literal2 = abjad.LilyPondLiteral(r"\once \override TextScript.self-alignment-X = 0")
        if abjad.get.indicator(self, literal1):
            abjad.detach(literal1, self)
        if abjad.get.indicator(self, literal2):
            abjad.detach(literal2, self)

    ### PUBLIC PROPERTIES ###

    @property
    def style(self) -> str:
        r"""The style of the harmonic note head."""
        return self._style

    @style.setter
    def style(
        self,
        style: str,
    ) -> None:
        if not isinstance(style, str):
            raise TypeError("'style' must be 'str'")
        self._style = style
        if not self._style.endswith("flageolet"):
            abjad.tweak(self._note_head).style = self._style
        else:
            flageolet = abjad.LilyPondLiteral(
                r"\flageolet",
                format_slot="after",
            )
            abjad.attach(flageolet, self)

    @property
    def centre_markup(self) -> bool:
        r"""Tweaks the markup of the harmonic note head to be centred or not
        as LilyPond doesn't centralises markups above note heads by default.
        """
        return self._centre_markup

    @centre_markup.setter
    def centre_markup(
        self,
        centre_markup: bool,
    ) -> None:
        if not isinstance(centre_markup, bool):
            raise TypeError("'style' must be 'bool'")
        self._centre_markup = centre_markup

    @property
    def markup(self) -> str:
        r"""The markup of the harmonic note head."""
        return self._markup

    @markup.setter
    def markup(
        self,
        markup: str,
    ) -> None:
        if markup is not None:
            if not isinstance(markup, str):
                raise TypeError("'markup' must be 'str'")
            self._markup = markup
            markup = abjad.Markup(
                self._markup,
                direction=self._direction,
            )
            abjad.attach(markup, self)
            if self._centre_markup:
                self._attach_centre_markup()
            else:
                self._detach_centre_markup()
        else:
            self._markup = markup
            if abjad.get.indicator(self, abjad.Markup):
                abjad.detach(abjad.Markup, self)
            self._detach_centre_markup()



from typing import Optional, Union

import abjad


def remove_repeated_dynamics(
    selection: abjad.Selection,
    *,
    ignore_hairpins: bool = False,
    reset_after_rests: bool = False,
    reset_after_duration: Optional[Union[float, int, str, tuple, abjad.Duration]] = None,
) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function removes all consecutive repeated dynamic markings.

    Basic usage:
        When two consecutive leaves have identical dynamics, the second
        one is removed:

        >>> staff = abjad.Staff(r"\time 3/8 c'4\pp d'8\pp | c'4\f d'8\f")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4
                \pp
                d'8
                \pp
                c'4
                \f
                d'8
                \f
            }

        ..  figure:: ../_images/remove_repeated_dynamics-anw32e9i0f.png

        >>> auxjad.mutate.remove_repeated_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4
                \pp
                d'8
                c'4
                \f
                d'8
            }

        ..  figure:: ../_images/remove_repeated_dynamics-i5ylxkzv7md.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.remove_repeated_dynamics(staff[:])
        >>> abjad.mutate.remove_repeated_dynamics(staff[:])

    Dynamic structure:
        The function also removes dynamics that are separated by an arbitrary
        number of leaves without dynamics:

        >>> staff = abjad.Staff(r"\time 3/8 c'4\p d'8 | e'4.\p | c'4\p d'8\f")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4
                \p
                d'8
                e'4.
                \p
                c'4
                \p
                d'8
                \f
            }

        ..  figure:: ../_images/remove_repeated_dynamics-ha1x7s8d2fb.png

        >>> auxjad.mutate.remove_repeated_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4
                \p
                d'8
                e'4.
                c'4
                d'8
                \f
            }

        ..  figure:: ../_images/remove_repeated_dynamics-g157jbbojhv.png

    Subcontainers:
        The container from which the selection is made can also have
        subcontainers:

        >>> staff = abjad.Staff([abjad.Note("c'2"),
        ...                      abjad.Chord("<d' f'>2"),
        ...                      abjad.Tuplet((2, 3), "g2 a2 b2"),
        ...                      ])
        >>> abjad.attach(abjad.Dynamic('ppp'), staff[0])
        >>> abjad.attach(abjad.Dynamic('ppp'), staff[1])
        >>> abjad.attach(abjad.Dynamic('ppp'), staff[2][0])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'2
                \ppp
                <d' f'>2
                \ppp
                \times 2/3
                {
                    g2
                    \ppp
                    a2
                    b2
                }
            }

        ..  figure:: ../_images/remove_repeated_dynamics-4h9xze4780d.png

        >>> auxjad.mutate.remove_repeated_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'2
                \ppp
                <d' f'>2
                \times 2/3
                {
                    g2
                    a2
                    b2
                }
            }

        ..  figure:: ../_images/remove_repeated_dynamics-7n9aaveoslu.png

    ``ignore_hairpins``:
        By default, repeated dynamics with hairpins in between are not removed,
        but consecutive ones will.

        >>> staff = abjad.Staff(r"c'2\p\< d'2\f\> | c'2\f d'2\f | e'1\p")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                d'1
                \f
                \>
                c'1
                \f
                d'1
                \f
                e'1
                \p
            }

        ..  figure:: ../_images/remove_repeated_dynamics-frmlobo3gis.png

        >>> auxjad.mutate.remove_repeated_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                d'1
                \f
                \>
                c'1
                \f
                d'1
                e'1
                \p
            }

        ..  figure:: ../_images/remove_repeated_dynamics-ov05k1imubj.png

        To override the previous behaviour, set ``ignore_hairpins=True`` and
        hairpins will be ignored.

        >>> staff = abjad.Staff(r"c'2\p\< d'2\f\> | c'2\f d'2\f | e'1\p")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                d'1
                \f
                \>
                c'1
                \f
                d'1
                \f
                e'1
                \p
            }

        ..  figure:: ../_images/remove_repeated_dynamics-2hdkt6cyca1.png

        >>> auxjad.mutate.remove_repeated_dynamics(
        ...     staff[:],
        ...     ignore_hairpins=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                d'1
                \f
                \>
                c'1
                d'1
                e'1
                \p
            }

        ..  figure:: ../_images/remove_repeated_dynamics-xkaipizr2jr.png

    ``reset_after_rests``:
        By default, rests are treated just like any other leaf and thus notes
        with an identical dynamic separated by an arbitrary number of rests
        will be considered as repeated and the second dynamic will be removed.

        >>> staff = abjad.Staff(r"c'4\pp r2. | c'1\pp")
        >>> auxjad.mutate.remove_repeated_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                r2.
                c'1
            }

        ..  figure:: ../_images/remove_repeated_dynamics-wtno2t8qroh.png

        To override the previous behaviour, set ``reset_after_rests`` to
        ``True`` and dynamics will always be restated after a rest.

        >>> staff = abjad.Staff(r"c'4\pp r2. | c'1\pp")
        >>> auxjad.mutate.remove_repeated_dynamics(
        ...     staff[:],
        ...     reset_after_rests=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                r2.
                c'1
                \pp
            }

        ..  figure:: ../_images/remove_repeated_dynamics-3e6g7u0q1i1.png

    ``reset_after_duration``:
        With ``reset_after_rests`` is set to ``True``, it is possible to
        specify a minimum duration of silence required for the dynamics to be
        restated using the argument ``reset_after_duration``. It takes an
        |abjad.Duration| or also :obj:`str`, :obj:`tuple`, :obj:`float`, etc.
        This sets the maximum length of rests before which identical dynamics
        are restated. If the total length of rests falls below that value, then
        repeated dynamics are removed.

        In the case below, a rest of ``r2``. is shorter than a duration of
        ``(4, 4)``, so the repeated dynamic is removed.

        >>> staff = abjad.Staff(r"c'4\pp r2. | c'1\pp")
        >>> auxjad.mutate.remove_repeated_dynamics(
        ...     staff[:],
        ...     reset_after_rests=True,
        ...     reset_after_duration=(4, 4),
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                r2.
                c'1
            }

        ..  figure:: ../_images/remove_repeated_dynamics-b323xuesujc.png

        But setting the duration to ``2/4`` forces the dynamic to be restated.

        >>> staff = abjad.Staff(r"c'4\pp r2. | c'1\pp")
        >>> auxjad.mutate.remove_repeated_dynamics(
        ...     staff[:],
        ...     reset_after_rests=True,
        ...     reset_after_duration=2 / 4,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                r2.
                c'1
                \pp
            }

        ..  figure:: ../_images/remove_repeated_dynamics-64dppx3cp99.png

    ``reset_after_rests``:
        The function also handles measure rests with ``reset_after_rests``.

        >>> staff = abjad.Staff(r"c'4\pp r2. | c'4\pp r2. | R1 | c'1\pp")
        >>> auxjad.mutate.remove_repeated_dynamics(
        ...     staff[:],
        ...     reset_after_rests=True,
        ...     reset_after_duration=abjad.Duration(4, 4),
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                r2.
                c'4
                r2.
                R1
                c'1
                \pp
            }

        ..  figure:: ../_images/remove_repeated_dynamics-jt7akhtbsge.png

    ..  warning::

        The input selection must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")
    if not abjad.select(selection).leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")
    if not isinstance(ignore_hairpins, bool):
        raise TypeError("'ignore_hairpins' must be 'bool'")
    if not isinstance(reset_after_rests, bool):
        raise TypeError("'reset_after_rests' must be 'bool'")
    if reset_after_duration is not None:
        if not isinstance(
            reset_after_duration,
            (
                abjad.Duration,
                str,
                tuple,
                int,
                float,
            ),
        ):
            raise TypeError(
                "'reset_after_duration' must be 'abjad.Duration', 'str', 'tuple', or a number"
            )
        if not isinstance(reset_after_duration, abjad.Duration):
            reset_after_duration = abjad.Duration(reset_after_duration)

    previous_dynamic = None
    current_dynamic = None
    duration_since_last_note = abjad.Duration(0)
    for leaf in selection.leaves():
        if isinstance(leaf, (abjad.Rest, abjad.MultimeasureRest)):
            dynamic = abjad.get.indicator(leaf, abjad.Dynamic)
            if dynamic is not None:
                previous_dynamic = dynamic
            if reset_after_rests:
                if reset_after_duration is None:
                    previous_dynamic = None
                else:
                    duration_since_last_note += leaf.written_duration
                    if duration_since_last_note >= reset_after_duration:
                        previous_dynamic = None
        else:
            duration_since_last_note = abjad.Duration(0)
            indicators = abjad.get.indicators(leaf)
            hairpin_present = any(
                [isinstance(indicator, abjad.StartHairpin) for indicator in indicators]
            )
            if hairpin_present and not ignore_hairpins:
                for indicator in indicators:
                    if isinstance(indicator, abjad.Dynamic):
                        if indicator is not None and indicator == previous_dynamic:
                            abjad.detach(abjad.Dynamic, leaf)
                current_dynamic = None
            else:
                for indicator in indicators:
                    if isinstance(indicator, abjad.Dynamic):
                        current_dynamic = indicator
            if current_dynamic != previous_dynamic:
                previous_dynamic = current_dynamic
            elif current_dynamic is not None:
                abjad.detach(abjad.Dynamic, leaf)



import abjad


def merge_partial_tuplets(
    selection: abjad.Selection,
    *,
    merge_across_barlines: bool = False,
) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function merges all consecutive partial tuplets with the same ratio
    and which sum up to an assignable duration. Partial tuplets can result from
    algorithmic manipulations such as phasing or looping, which can slice
    through a tuplet.

    Basic usage:
        Usage is simple:

        >>> staff = abjad.Staff(r"\times 2/3 {c'1} \times 2/3 {d'2}")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'1
                }
                \times 2/3
                {
                    d'2
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-ilr68s15kqb.png

        >>> auxjad.mutate.merge_partial_tuplets(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'1
                    d'2
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-qe29etsedx.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.merge_partial_tuplets(staff[:])
        >>> abjad.mutate.merge_partial_tuplets(staff[:])

    Multiple consecutive partial tuplets:
        This function can also handle several consecutive partial tuplets:

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {c'2} \times 2/3 {d'2} \times 2/3 {e'2}"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            {
                \times 2/3
                {
                    c'2
                }
                \times 2/3
                {
                    d'2
                }
                \times 2/3
                {
                    e'2
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-9rh7vpu208j.png

        >>> auxjad.mutate.merge_partial_tuplets(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'2
                    d'2
                    e'2
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-oy1imqisx2.png

    ``merge_across_barlines``:
        By default, partial tuplets are not merged across barlines.

        >>> staff = abjad.Staff(
        ...     r"\time 3/4 c'2. "
        ...     r"\times 2/3 {d'4} r4 \times 2/3 {e'2} "
        ...     r"\times 2/3 {f'4} r4 \times 2/3 {g'2}"
        ... )
        >>> auxjad.mutate.merge_partial_tuplets(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    d'4
                }
                r4
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    e'2
                }
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    f'4
                }
                r4
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    g'2
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-3rjib7pctml.png

        To change  this behaviour, set ``merge_across_barlines`` to ``True``.

        >>> staff = abjad.Staff(
        ...     r"\time 3/4 c'2. "
        ...     r"\times 2/3 {d'4} r4 \times 2/3 {e'2} "
        ...     r"\times 2/3 {f'4} r4 \times 2/3 {g'2}"
        ... )
        >>> auxjad.mutate.merge_partial_tuplets(
        ...     staff[:],
        ...     merge_across_barlines=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    d'4
                }
                r4
                \times 2/3
                {
                    e'2
                    f'4
                }
                r4
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    g'2
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-icud1ejcmzc.png

    Tied partial tuplets:
        Tied partial tuplets are also handled by this function.

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {r4} \times 2/3 {c'2} "
        ...     r"\times 4/5 {d'2~} \times 4/5{d'8}"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    r4
                }
                \times 2/3
                {
                    c'2
                }
                \times 4/5
                {
                    d'2
                    ~
                }
                \times 4/5
                {
                    d'8
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-st4zw38qfce.png

        >>> auxjad.mutate.merge_partial_tuplets(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    r4
                    c'2
                }
                \times 4/5
                {
                    d'2
                    ~
                    d'8
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-1pky5fsh2nl.png

    Indicators:
        Indicators stay the same in the merged tuplet.

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {c'2\p\< d'2} \times 2/3 {e'2\ff}"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'2
                    \p
                    \<
                    d'2
                }
                \times 2/3
                {
                    e'2
                    \ff
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-7cdtafl348h.png

        >>> auxjad.mutate.merge_partial_tuplets(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'2
                    \p
                    \<
                    d'2
                    e'2
                    \ff
                }
            }

        ..  figure:: ../_images/merge_partial_tuplets-j9rmdfbawce.png

    ..  tip::

        The method |auxjad.mutate.extract_trivial_tuplets()| can be used
        after merging partial tuplets to further clean the output. The method
        |auxjad.mutate.auto_rewrite_meter()| can also be used for this
        purpose, as it will not only rewrite the metric notation of a staff but
        also apply both |auxjad.mutate.merge_partial_tuplets()| and
        |auxjad.mutate.extract_trivial_tuplets()| to the output.

    ..  note::

        When using |abjad.Container|'s, all time signatures in the output will
        be commented out with ``%%%.`` This is because Abjad only applies time
        signatures to containers that belong to a |abjad.Staff|. The present
        function works with either |abjad.Container| and |abjad.Staff|.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> abjad.show(container)

        ..  docs::

            {
                %%% \time 3/4 %%%
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/merge_partial_tuplets-945s36mfdn.png

        >>> staff = abjad.Staff([container])
        >>> abjad.show(container)

        ..  docs::

            {
                \time 3/4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/merge_partial_tuplets-3b4tyqrnttw.png

    ..  warning::

        The input selection must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")
    if not isinstance(merge_across_barlines, bool):
        raise TypeError("'merge_across_barlines' must be 'bool'")

    tuplets = selection.tuplets()
    if len(tuplets) <= 1:
        return

    def _process_tuplets(tuplets):
        for index in range(len(tuplets[:-1])):
            for upper_index in range(index, len(tuplets)):
                if tuplets[index].multiplier == tuplets[upper_index].multiplier:
                    tuplet_group = tuplets[index : upper_index + 1]
                else:
                    break
                if tuplet_group.are_contiguous_logical_voice():
                    durations = [abjad.get.duration(tuplet) for tuplet in tuplet_group]
                    sum_durations = sum(durations)
                    if (
                        all(duration.implied_prolation != 1 for duration in durations)
                        and sum_durations.implied_prolation == 1
                    ):
                        for tuplet in tuplet_group[1:]:
                            tuplet_group[0].extend(tuplet)
                            abjad.mutate.extract(tuplet)

    if not merge_across_barlines:
        measures = selection.group_by_measure()
        for measure in measures:
            tuplets = abjad.select(measure).tuplets()
            if len(tuplets) <= 1:
                continue
            _process_tuplets(tuplets)
    else:
        _process_tuplets(tuplets)



import abjad


def remove_repeated_time_signatures(selection: abjad.Selection) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function removes all consecutive repeated time signatures.

    Basic usage:
        When two consecutive measures have identical time signatures, the
        second one is removed:

        >>> staff = abjad.Staff(r"c'4 d'8 | c'4 d'8")
        >>> abjad.attach(abjad.TimeSignature((3, 8)), staff[0])
        >>> abjad.attach(abjad.TimeSignature((3, 8)), staff[2])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4
                d'8
                \time 3/8
                c'4
                d'8
            }

        ..  figure:: ../_images/remove_repeated_time_signatures-feZSi4Trsg.png

        >>> auxjad.mutate.remove_repeated_time_signatures(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4
                d'8
                c'4
                d'8
            }

        ..  figure:: ../_images/remove_repeated_time_signatures-ImmpJOWn5U.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.remove_repeated_time_signatures(staff[:])
        >>> abjad.mutate.remove_repeated_time_signatures(staff[:])

    Time signature structure:
        The function also removes time signatures that are separated by an
        arbitrary number of measures without one:

        >>> staff = abjad.Staff(r"c'4 d'8 e'4. c'4 d'8")
        >>> abjad.attach(abjad.TimeSignature((3, 8)), staff[0])
        >>> abjad.attach(abjad.TimeSignature((3, 8)), staff[3])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4
                d'8
                e'4.
                \time 3/8
                c'4
                d'8
            }

        ..  figure:: ../_images/remove_repeated_time_signatures-ihs4kU1dMe.png

        >>> auxjad.mutate.remove_repeated_time_signatures(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4
                d'8
                e'4.
                c'4
                d'8
            }

        ..  figure:: ../_images/remove_repeated_time_signatures-Ans1RrG5ZW.png

    Subcontainers:
        The container from which the selection is made can also have
        subcontainers, including cases in which the time signatures are
        attached to leaves of subcontainers:

        >>> staff = abjad.Staff([abjad.Note("c'2"),
        ...                      abjad.Chord("<d' f'>2"),
        ...                      abjad.Tuplet((2, 3), "g2 a2 b2"),
        ...                      ])
        >>> abjad.attach(abjad.TimeSignature((2, 2)), staff[0])
        >>> abjad.attach(abjad.TimeSignature((2, 2)), staff[2][0])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/2
                c'2
                <d' f'>2
                \times 2/3
                {
                    \time 2/2
                    g2
                    a2
                    b2
                }
            }

        ..  figure:: ../_images/remove_repeated_time_signatures-Nybwh816FT.png

        >>> auxjad.mutate.remove_repeated_time_signatures(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/2
                c'2
                <d' f'>2
                \times 2/3
                {
                    g2
                    a2
                    b2
                }
            }

        ..  figure:: ../_images/remove_repeated_time_signatures-PNCfPcnTtj.png

    ..  warning::

        The input selection must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")
    if not selection.leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")

    measures = selection.group_by_measure()
    head = selection.leaf(0)
    previous_time_signature = abjad.get.indicator(
        head,
        abjad.TimeSignature,
    )
    if previous_time_signature is None:
        previous_time_signature = abjad.TimeSignature((4, 4))
    for measure in measures[1:]:
        head = abjad.select(measure).leaf(0)
        time_signature = abjad.get.indicator(head, abjad.TimeSignature)
        if time_signature == previous_time_signature:
            abjad.detach(abjad.TimeSignature, head)
        elif time_signature is not None:
            previous_time_signature = time_signature



from collections.abc import Iterable
from typing import Union

import abjad

from .. import select


def rhythms_are_identical(
    selections: Union[
        Iterable[abjad.Component],
        Iterable[abjad.Selection],
    ],
) -> bool:
    r"""Returns a :obj:`bool` representing whether two or more selections are
    rhytmically identical or not. Input argument must be an iterable made of
    two or more |abjad.Selection|'s.

    Basic usage:
        When the pitches and effective durations of all leaves in all
        selections are identical, this function returns ``True``:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Staff(r"c''4 b'4 a'4 g'4")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.rhythms_are_identical(selections)
        True

        Rests and chords are also handled.

        >>> container1 = abjad.Staff(r"c'4. d'8 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"c''4. r8 <b' a'>4 r4 g'2 f'2")

        When comparing rests, this function considers consecutive rests as a
        single logical tie.

        >>> container1 = abjad.Staff(
        ...     r"c'4 ~ c'16 r8. r8. <d' e'>16 ~ <d' e'>4"
        ... )
        >>> container2 = abjad.Staff(r"r4 r16 c''8. ~ c''8. b'16 ~ b'4")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.rhythms_are_identical(selections)
        True
        >>> container1 = abjad.Staff(
        ...     r"c'4 ~ c'16 r8. r8. <d' e'>16 ~ <d' e'>4"
        ... )
        >>> container2 = abjad.Staff(r"r4 r16 c''8. b'8. a'16 ~ a'4")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.rhythms_are_identical(selections)
        False

        This function can handle multiple selections, and will compare them
        among each other returning ``True`` if all are identical:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"c''4 r4 <b' a'>4 r4 g'2 f'2")
        >>> container3 = abjad.Staff(r"c'''4 c'''4 b''4 b''4 a''2 a''2")
        >>> container4 = abjad.Staff(r"c4 d4 e4 f4 g2 a2")
        >>> selections = [
        ...     container1[:],
        ...     container2[:],
        ...     container3[:],
        ...     container4[:],
        ... ]
        >>> auxjad.get.rhythms_are_identical(selections)
        True

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.get|. It can thus be used from either |auxjad.get|_ or
        |abjad.get| namespaces. Therefore, the two lines below are equivalent:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Staff(r"c''4 b'4 a'4 g'4")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.rhythms_are_identical(selections)
        True
        >>> abjad.get.rhythms_are_identical(selections)
        True

    Effective durations:
        Even if all leaves of both selections are identical in relation to both
        pitches and written durations, the function considers the effective
        durations. This means that situations like the one below do not yield a
        false positive:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(
        ...     r"\times 3/2 {c'4 d'4 e'4} f'4 <g' a'>2 r2"
        ... )
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.rhythms_are_identical(selections)
        False

    Grace notes:
        This function also handles grace notes.

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Staff(r"c'4 \grace{d'4} d'4 e'4 f'4")
        >>> selection1 = abjad.select(container1)
        >>> selection2 = abjad.select(container2)
        >>> selections = [selection1, selection2]
        >>> auxjad.get.rhythms_are_identical(selections)
        False

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(
        ...     r"c'4 \grace{c''4} d'4 e'4 f'4 <g' a'>2 r2"
        ... )
        >>> selection1 = abjad.select(container1)
        >>> selection2 = abjad.select(container2)
        >>> selections = [selection1, selection2]
        >>> auxjad.get.rhythms_are_identical(selections)
        False

        >>> container1 = abjad.Staff(
        ...     r"c'4 \grace{c''4} d'4 e'4 f'4 <g' a'>2 r2"
        ... )
        >>> container2 = abjad.Staff(
        ...     r"c'4 \grace{b4} d'4 e'4 f'4 <g' a'>2 r2"
        ... )
        >>> selection1 = abjad.select(container1)
        >>> selection2 = abjad.select(container2)
        >>> selections = [selection1, selection2]
        >>> auxjad.get.rhythms_are_identical(selections)
        True

    ..  warning::

        It is important to create selections using |abjad.select()| as shown in
        the example above, instead of using the syntax ``container[:]``, since
        the latter ignores grace notes.

    ..  note::

        It is important to note it is the contents of the containers which are
        compared, so containers of different classes can still return a
        ``True`` value.

        >>> container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.rhythms_are_identical(selections)
        True
    """
    if not isinstance(selections, Iterable):
        raise TypeError("argument must be an iterable of 'abjad.Selection's or 'abjad.Component's")
    for selection in selections:
        if not isinstance(selection, (abjad.Component, abjad.Selection)):
            raise TypeError(
                "argument must be an iterable of 'abjad.Selection's or 'abjad.Component's"
            )

    for index, selection1 in enumerate(selections[:-1]):
        for selection2 in selections[index + 1 :]:
            logical_selections1 = [
                logical_selection for logical_selection in select.logical_selections(selection1)
            ]
            logical_selections2 = [
                logical_selection for logical_selection in select.logical_selections(selection2)
            ]
            if len(logical_selections1) != len(logical_selections2):
                return False
            for logical_sel1, logical_sel2 in zip(
                logical_selections1,
                logical_selections2,
            ):
                if abjad.get.duration(logical_sel1) != abjad.get.duration(logical_sel2):
                    return False
                leaf1 = logical_sel1.leaf(0)
                leaf2 = logical_sel2.leaf(0)
                leaf1_graces = abjad.get.before_grace_container(leaf1)
                leaf2_graces = abjad.get.before_grace_container(leaf2)
                if not isinstance(leaf1_graces, type(leaf2_graces)):
                    return False
    return True



from math import ceil
from typing import Optional, Union

import abjad

from .. import mutate
from ._LooperParent import _LooperParent


class WindowLooper(_LooperParent):
    r"""Outputs slices of an |abjad.Container| (or child class) using the
    metaphor of a looping window of a constant size (given by an
    |abjad.Duration|).

    Basic usage:
        Calling the object will return an |abjad.Selection| generated by the
        looping process. Each call of the object will move the window forwards
        and output the sliced window. It requires an :attr:`window_size` and a
        :attr:`step_size`.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                d'2
                e'4
                f'2
                ~
                f'8
                g'4.
            }

        ..  figure:: ../_images/WindowLooper-trzwtEJ4uG.png

        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 16),
        ...                              )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'2
                e'4
            }

        ..  figure:: ../_images/WindowLooper-a9k9q8xy1j.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                d'16
                ~
                d'4
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
            }

        ..  figure:: ../_images/WindowLooper-oori8gjer9s.png

        The property :attr:`current_window` can be used to access the current
        window without moving the head forwards.

        >>> notes = looper.current_window()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                d'16
                ~
                d'4
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
            }

        ..  figure:: ../_images/WindowLooper-7zc9e7o3dlr.png

    :attr:`process_on_first_call`:
        The very first call will output the input container without processing
        it. To disable this behaviour and have the looping window move on the
        very first call, initialise the class with the keyword argument
        :attr:`process_on_first_call` set to ``True``.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 16),
        ...                              process_on_first_call=True,
        ...                              )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                d'16
                ~
                d'4
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
            }

        ..  figure:: ../_images/WindowLooper-dual73tnheq.png

    :attr:`window_size` and :attr:`step_size`:
        The arguments :attr:`window_size` and :attr:`step_size` are used to set
        the sizes of the window and the process step. :attr:`window_size` can
        take a :obj:`tuple` or an |abjad.Meter| as input, while
        :attr:`step_size` takes a :obj:`tuple` or an |abjad.Duration|.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 8),
        ...                              )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'2
            }

        ..  figure:: ../_images/WindowLooper-0wh7ajyal0qj.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'8
                d'8
                ~
                d'4.
                e'8
            }

        ..  figure:: ../_images/WindowLooper-tt90u6gg2tp.png

    Using as iterator:
        The instances of this class can also be used as an iterator, which can
        then be used in a for loop to exhaust all windows. Notice how it
        appends rests at the end of the container, until it is totally
        exhausted. Note that unlike the methods :meth:`output_n` and
        :meth:`output_all`, time signatures are added to each window returned
        by the shuffler. Use the function
        |auxjad.mutate.remove_repeated_time_signatures()| to clean the output
        when using this class in this way.

        >>> container = abjad.Container(r"c'4 d'2 e'4")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 8),
        ...                              )
        >>> staff = abjad.Staff()
        >>> for window in looper:
        ...     staff.append(window)
        >>> auxjad.mutate.remove_repeated_time_signatures(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'2
                c'8
                d'8
                ~
                d'4.
                e'8
                d'2
                e'4
                d'4.
                e'4
                r8
                d'4
                e'4
                r4
                d'8
                e'4
                r4.
                e'4
                r2
                e'8
                r8
                r2
            }

        ..  figure:: ../_images/WindowLooper-9sldax4dumb.png

    :attr:`fill_with_rests`:
        In order to stop the process when the end of the looping window matches
        the end of the :attr:`contents` (and thus appending rests to the
        output), set the optional keyword argument :attr:`fill_with_rests` to
        ``True``. Compare the two approaches below.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> staff = abjad.Staff()
        >>> for window in looper:
        ...     staff.append(window)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                d'4
                e'4
                f'4
                e'4
                f'4
                r4
                f'4
                r2
            }

        ..  figure:: ../_images/WindowLooper-4l025g7ycxr.png

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              fill_with_rests=False,
        ...                              )
        >>> staff = abjad.Staff()
        >>> for window in looper:
        ...     staff.append(window)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/WindowLooper-lq4cwlzlpp.png

    Arguments and properties:
        This class can take many optional keyword arguments during its
        creation. :attr:`max_steps` sets the maximum number of steps that the
        window can advance when the object is called, ranging between ``1`` and
        the input value (default is also ``1``). :attr:`repetition_chance` sets
        the chance of a window result repeating itself (that is, the window not
        moving forwards when called). It should range from ``0.0`` to ``1.0``
        (default ``0.0``, i.e. no repetition). :attr:`forward_bias` sets the
        chance of the window moving forward instead of backwards. It should
        range from ``0.0`` to ``1.0`` (default ``1.0``, which means the window
        can only move forwards. A value of ``0.5`` gives 50% chance of moving
        forwards while a value of ``0.0`` will move the window only backwards).
        :attr:`head_position` can be used to offset the starting position of
        the looping window. It must be a :obj:`tuple` or an |abjad.Duration|,
        and its default value is ``0``. The properties :attr:`boundary_depth`,
        :attr:`maximum_dot_count`, and :attr:`rewrite_tuplets` are passed as
        arguments to |abjad.Meter.rewrite_meter()|, see its documentation
        for more information. By default, calling the object will first return
        the original container and subsequent calls will process it; set
        :attr:`process_on_first_call` to ``True`` and the looping process will
        be applied on the very first call. The attribute :attr:`after_rest`
        sets the length of rests that separate consecutive iterations (default
        is ``abjad.Duration(0)``, i.e. no rest). When using after rests,
        :attr:`after_rest_in_new_measure` dictates whether or not rests are
        appended to the same output measure or to a new one. When in their own
        measure, :attr:`use_multimeasure_rests` sets if multi-measure rests
        should be used or not.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(5, 8),
        ...                              max_steps=2,
        ...                              repetition_chance=0.25,
        ...                              forward_bias=0.2,
        ...                              head_position=(2, 8),
        ...                              omit_time_signatures=False,
        ...                              fill_with_rests=False,
        ...                              boundary_depth=0,
        ...                              maximum_dot_count=1,
        ...                              rewrite_tuplets=False,
        ...                              process_on_first_call=True,
        ...                              after_rest=(1, 8),
        ...                              after_rest_in_new_measure=True,
        ...                              use_multimeasure_rests=False,
        ...                              )
        >>> looper.window_size
        3/4
        >>> looper.step_size
        5/8
        >>> looper.repetition_chance
        0.25
        >>> looper.forward_bias
        0.2
        >>> looper.max_steps
        2
        >>> looper.head_position
        1/4
        >>> looper.omit_time_signatures
        False
        >>> looper.fill_with_rests
        False
        >>> looper.boundary_depth
        0
        >>> looper.maximum_dot_count
        1
        >>> looper.rewrite_tuplets
        False
        >>> looper.boundary_depth
        0
        >>> looper.maximum_dot_count
        1
        >>> looper.rewrite_tuplets
        False
        >>> looper.process_on_first_call
        True
        >>> looper.after_rest
        abjad.Duration((1, 8))
        >>> looper.after_rest_in_new_measure
        True
        >>> looper.use_multimeasure_rests
        False

        Use the properties below to change these values after initialisation.

        >>> looper.window_size = (5, 4)
        >>> looper.step_size = (1, 4)
        >>> looper.max_steps = 3
        >>> looper.repetition_chance = 0.1
        >>> looper.forward_bias = 0.8
        >>> looper.head_position = 0
        >>> looper.omit_time_signatures = True
        >>> looper.boundary_depth = 1
        >>> looper.maximum_dot_count = 2
        >>> looper.rewrite_tuplets = True
        >>> looper.process_on_first_call = False
        >>> looper.after_rest = 0
        >>> looper.after_rest_in_new_measure = False
        >>> looper.use_multimeasure_rests = True
        >>> looper.window_size
        5/4
        >>> looper.step_size
        1/4
        >>> looper.max_steps
        3
        >>> looper.repetition_chance
        0.1
        >>> looper.forward_bias
        0.8
        >>> looper.head_position
        0
        >>> looper.omit_time_signatures
        True
        >>> looper.boundary_depth
        1
        >>> looper.maximum_dot_count
        2
        >>> looper.rewrite_tuplets
        True
        >>> looper.process_on_first_call
        False
        >>> looper.after_rest
        abjad.Duration(0)
        >>> looper.after_rest_in_new_measure
        False
        >>> looper.use_multimeasure_rests
        True

    Setting :attr:`forward_bias` to ``0.0``:
        Set :attr:`forward_bias` to ``0.0`` to move backwards instead of
        forwards (default is ``1.0``). The initial :attr:`head_position` must
        be greater than ``0`` otherwise the contents will already be exhausted
        in the very first call (since it will not be able to move backwards
        from that position).

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              head_position=(3, 4),
        ...                              forward_bias=0.0,
        ...                              )
        >>> notes = looper.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                f'4
                g'4
                a'4
                e'4
                f'4
                g'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/WindowLooper-ty5vqw6qk9d.png

    :attr:`forward_bias` between ``0.0`` and ``1.0``:
        Setingt :attr:`forward_bias` to a value in between ``0.0`` and ``1.0``
        will result in random steps being taken forward or backward, according
        to the bias. The initial value of :attr:`head_position` will once gain
        play an important role here, as the contents might be exhausted if the
        looper attempts to move backwards after reaching the head position
        ``0``.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4 b'4 c''4")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              head_position=(3, 4),
        ...                              forward_bias=0.5,
        ...                              )
        >>> notes = looper.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                f'4
                g'4
                a'4
                e'4
                f'4
                g'4
                d'4
                e'4
                f'4
                e'4
                f'4
                g'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/WindowLooper-p91xgsnmp3o.png

    :attr:`max_steps`:
        Setting the keyword argument :attr:`max_steps` to a value larger than
        ``1`` will result in a random number of steps (between ``1`` and
        :attr:`max_steps`) being applied at each call.

        >>> container = abjad.Container(
        ...     r"c'4 d'4 e'4 f'4 g'4 a'4 b'4 c''4 d''4 e''4 f''4"
        ... )
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              max_steps=4,
        ...                              )
        >>> notes = looper.output_n(4)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                e'4
                f'4
                g'4
                f'4
                g'4
                a'4
                c''4
                d''4
                e''4
            }

        ..  figure:: ../_images/WindowLooper-81mbjuesmbr.png

    :func:`len()`:
        The function :func:`len()` can be used to get the total number of steps
        in the contents (always rounded up).

        >>> container = abjad.Container(r"c'1")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 16),
        ...                              )
        >>> len(looper)
        16
        >>> container = abjad.Container(r"c'1")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> len(looper)
        4
        >>> container = abjad.Container(r"c'2..")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(2, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> len(looper)
        4

    :meth:`output_all`:
        To run through the whole process and output it as a single container,
        from the initial head position until the process outputs the single
        last element, use the method :meth:`output_all`. As shown above, set
        the optional keyword argument :attr:`fill_with_rests` to ``False`` if
        the process is to be stopped when the end of the looping window reaches
        the end of the contents (thus not appending rests).

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> notes = looper.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                d'4
                e'4
                f'4
                e'4
                f'4
                r4
                f'4
                r2
            }

        ..  figure:: ../_images/WindowLooper-y734t07uio.png

    ``tie_identical_pitches``:
        When using :meth:`output_all`, set the keyword argument
        ``tie_identical_pitches`` to ``True`` in order to tie identical notes
        or chords at the end and beginning of consecutive windows.

        >>> container = abjad.Container(r"c'4 <e' f' g'>2 r4 f'2.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> notes = looper.output_all(tie_identical_pitches=True)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                <e' f' g'>2
                ~
                <e' f' g'>2
                r4
                <e' f' g'>4
                r4
                f'4
                r4
                f'2
                ~
                f'2.
                ~
                f'2
                r4
                f'4
                r2
            }

        ..  figure:: ../_images/WindowLooper-1giyp118geth.png

    :meth:`output_n`:
        To run through just part of the process and output it as a single
        container, starting from the initial head position, use the method
        :meth:`output_n` and pass the number of iterations as argument.
        Similarly to :meth:`output_all`, the optional keyword arguments
        ``tie_identical_pitches`` and :attr:`fill_with_rests` are available.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 4),
        ...                              )
        >>> notes = looper.output_n(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/WindowLooper-uikg5s4t26.png

    ..  tip::

        All methods that return an |abjad.Selection| will add an initial time
        signature to it. The :meth:`output_n` and :meth:`output_all` methods
        automatically remove repeated time signatures. When joining selections
        output by multiple method calls, use
        |auxjad.mutate.remove_repeated_time_signatures()| on the whole
        container after fusing the selections to remove any unecessary time
        signature changes.

    :attr:`omit_time_signatures`:
        To disable time signatures altogether, initialise this class with the
        keyword argument :attr:`omit_time_signatures` set to ``True`` (default
        is ``False``), or use the :attr:`omit_time_signatures` property after
        initialisation.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 16),
        ...                              omit_time_signatures=True,
        ...                              )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'2
                e'4
            }

        ..  figure:: ../_images/WindowLooper-24ipt4uf4x8.png

    :attr:`after_rest`:
        To append rests after the output (useful to create separations between
        consecutive windows), set the keyword argument :attr:`after_rest` to
        an :obj:`int`, :obj:`float`, :obj:`tuple`, or |abjad.Duration| with the
        total length of the rests. Setting it to ``abjad.Duration(0)`` will
        disable rests (which is the default value).

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 16),
        ...                              after_rest=(1, 4),
        ...                              )
        >>> notes = looper.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'2
                r4
                c'8.
                d'16
                ~
                d'4
                ~
                d'8.
                e'16
                r4
                c'8
                d'8
                ~
                d'4
                ~
                d'8
                e'8
                r4
            }

        ..  figure:: ../_images/WindowLooper-Oh9HqIcxyr.png

    :attr:`after_rest_in_new_measure`:
        When using after rests, the keyword argument
        :attr:`after_rest_in_new_measure` controls whether or not the rests are
        appended to the same output measure or to a new one.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 16),
        ...                              after_rest=(1, 4),
        ...                              after_rest_in_new_measure=True,
        ...                              )
        >>> notes = looper.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'2
                \time 1/4
                R1 * 1/4
                \time 3/4
                c'8.
                d'16
                ~
                d'4..
                e'16
                \time 1/4
                R1 * 1/4
                \time 3/4
                c'8
                d'8
                ~
                d'4.
                e'8
                \time 1/4
                R1 * 1/4
            }

        ..  figure:: ../_images/WindowLooper-PsrOr76UJu.png

    :attr:`use_multimeasure_rests`:
        When after rests are used and :attr:`after_rest_in_new_measure` is set
        to ``True``, multi-measure rests are automatically used. To use regular
        rests, set :attr:`use_multimeasure_rests` to ``False``.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 16),
        ...                              after_rest=(1, 4),
        ...                              after_rest_in_new_measure=True,
        ...                              use_multimeasure_rests=False,
        ...                              )
        >>> notes = looper.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'2
                \time 1/4
                r4
                \time 3/4
                c'8.
                d'16
                ~
                d'4..
                e'16
                \time 1/4
                r4
                \time 3/4
                c'8
                d'8
                ~
                d'4.
                e'8
                \time 1/4
                r4
            }

        ..  figure:: ../_images/WindowLooper-7Kg5fllsiV.png

    ..  note::

        Multi-measure rests will not be used if :attr:`omit_time_signatures` is
        ``True``, regardless of the value of :attr:`use_multimeasure_rests`.

    :attr:`window_size`:
        To change the size of the looping window after instantiation, use the
        property :attr:`window_size`. In the example below, the initial window
        is of size ``(4, 4)``, but changes to ``(3, 8)`` after three calls.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 16),
        ...                              )
        >>> notes = looper.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'2
                e'4
                c'8.
                d'16
                ~
                d'4
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
                c'8
                d'8
                ~
                d'4
                ~
                d'8
                e'4
                f'8
            }

        ..  figure:: ../_images/WindowLooper-52p5g0hqnep.png

        >>> looper.window_size = (3, 8)
        >>> notes = looper.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'16
                d'16
                ~
                d'4
                d'4.
                d'4.
            }

        ..  figure:: ../_images/WindowLooper-kzrb0fwup2.png

    :attr:`contents`:
        Use the :attr:`contents` property to read as well as overwrite the
        contents of the looper. Notice that the :attr:`head_position` will
        remain on its previous value and must be reset to ``0`` if that's
        required.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 16),
        ...                              )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'2
                e'4
            }

        ..  figure:: ../_images/WindowLooper-hx2wjkgko3j.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                d'16
                ~
                d'4
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
            }

        ..  figure:: ../_images/WindowLooper-bjabk27oapb.png

        >>> looper.contents = abjad.Container(r"c'16 d'16 e'16 f'16 g'2. a'1")
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                d'16
                e'16
                f'16
                g'16
                ~
                g'2
                ~
                g'8.
                a'16
            }

        ..  figure:: ../_images/WindowLooper-fdioda2e2ed.png

        >>> looper.head_position = 0
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'16
                d'16
                e'16
                f'16
                g'2.
            }

        ..  figure:: ../_images/WindowLooper-vh12q82nw0e.png

    Indicators:
        This class can handle dynamics, articulations and slurs. When a leaf is
        shortened by the looping window's movement, the dynamics and
        articulations are still applied to it.

        >>> container = abjad.Container(
        ...     r"c'4-.\p\< d'2--\f e'4->\ppp f'2 ~ f'8"
        ... )
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 16),
        ...                              )
        >>> notes = looper.output_n(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \p
                - \staccato
                \<
                d'2
                \f
                - \tenuto
                e'4
                \ppp
                - \accent
                c'8.
                \p
                - \staccato
                \<
                d'16
                \f
                - \tenuto
                ~
                d'4
                ~
                d'8.
                e'16
                \ppp
                - \accent
                ~
                e'8.
                f'16
            }

        ..  figure:: ../_images/WindowLooper-23jcqhmpqfm.png

    ..  tip::

        The functions |auxjad.mutate.remove_repeated_dynamics()| and
        |auxjad.mutate.reposition_clefs()| can be used to clean the output
        and remove repeated dynamics and unnecessary clef changes.

    ..  warning::

        Do note that some elements that span multiple notes (such as ottava
        indicators, manual beams, etc.) can become problematic when notes
        containing them are split into two. As a rule of thumb, it is always
        better to attach those to the music after the fading process has ended.

    :attr:`disable_rewrite_meter`:
        By default, this class uses the |abjad.Meter.rewrite_meter()|
        mutation.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4 g'4")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 16),
        ...                              )
        >>> notes = looper.output_n(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                c'8.
                d'16
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
            }

        ..  figure:: ../_images/LeafLooper-toNYz7MCq3.png

        Set :attr:`disable_rewrite_meter` to ``True`` in order to disable this
        behaviour.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4 g'4")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 16),
        ...                              disable_rewrite_meter=True,
        ...                              )
        >>> notes = looper.output_n(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                c'8.
                d'4
                e'4
                f'16
            }

        ..  figure:: ../_images/LeafLooper-osLcSeQ6gP.png

    Tweaking |abjad.Meter.rewrite_meter()|:
        This class uses the default logical tie splitting algorithm from
        |abjad.Meter.rewrite_meter()|.

        >>> container = abjad.Container(r"c'4. d'8 e'2")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 16),
        ...                              )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4.
                d'8
                e'2
            }

        ..  figure:: ../_images/WindowLooper-5iew8d28rqj.png

        Set :attr:`boundary_depth` to a different number to change its
        behaviour.

        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(4, 4),
        ...                              step_size=(1, 16),
        ...                              boundary_depth=1,
        ...                              )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                ~
                c'8
                d'8
                e'2
            }

        ..  figure:: ../_images/WindowLooper-hzetruc3kz4.png

        Other arguments available for tweaking the output of
        |abjad.Meter.rewrite_meter()| are :attr:`maximum_dot_count` and
        :attr:`rewrite_tuplets`, which work exactly as the identically named
        arguments of |abjad.Meter.rewrite_meter()|.

        This class also accepts the arguments ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, ``fuse_triple_meter``, and
        ``extract_trivial_tuplets``, which are passed on to
        |auxjad.mutate.prettify_rewrite_meter()| (the latter can be disabled
        by setting ``prettify_rewrite_meter`` to ``False``). See the
        documentation of this function for more details on these arguments.

    ..  warning::

        This class can handle tuplets, but the output is often quite complex.
        Although the result will be rhythmically correct, consecutive tuplets
        are not fused together, and tuplets may be output off-beat. This
        functionality should be considered experimental.

        >>> container = abjad.Container(r"\times 2/3 {c'8 d'8 e'8} d'2.")
        >>> looper = auxjad.WindowLooper(container,
        ...                              window_size=(3, 4),
        ...                              step_size=(1, 16),
        ...                              )
        >>> notes = looper.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    \time 3/4
                    c'8
                    d'8
                    e'8
                }
                d'2
                \times 2/3
                {
                    c'32
                    d'16
                    ~
                    d'16
                    e'8
                }
                d'16
                ~
                d'2
                \times 2/3
                {
                    d'16
                    e'8
                }
                d'8
                ~
                d'2
            }

        ..  figure:: ../_images/WindowLooper-cevvf9c9a9.png
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_omit_time_signatures",
        "_fill_with_rests",
        "_contents_length",
        "_contents_no_time_signature",
        "_disable_rewrite_meter",
        "_boundary_depth",
        "_maximum_dot_count",
        "_rewrite_tuplets",
        "_prettify_rewrite_meter",
        "_extract_trivial_tuplets",
        "_fuse_across_groups_of_beats",
        "_fuse_quadruple_meter",
        "_fuse_triple_meter",
        "_after_rest",
        "_after_rest_in_new_measure",
        "_use_multimeasure_rests",
    )

    ### INITIALISER ###

    def __init__(
        self,
        contents: abjad.Container,
        *,
        window_size: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
            abjad.Meter,
        ],
        step_size: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
        ],
        max_steps: int = 1,
        repetition_chance: float = 0.0,
        forward_bias: float = 1.0,
        head_position: Union[
            int,
            float,
            tuple[int],
            abjad.Duration,
        ] = 0,
        omit_time_signatures: bool = False,
        process_on_first_call: bool = False,
        fill_with_rests: bool = True,
        disable_rewrite_meter: bool = False,
        boundary_depth: Optional[int] = None,
        maximum_dot_count: Optional[int] = None,
        rewrite_tuplets: bool = True,
        prettify_rewrite_meter: bool = True,
        extract_trivial_tuplets: bool = True,
        fuse_across_groups_of_beats: bool = True,
        fuse_quadruple_meter: bool = True,
        fuse_triple_meter: bool = True,
        after_rest: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
            abjad.Rest,
        ] = 0,
        after_rest_in_new_measure: bool = False,
        use_multimeasure_rests: bool = True,
    ) -> None:
        self.contents = contents
        self.omit_time_signatures = omit_time_signatures
        self.fill_with_rests = fill_with_rests
        self.disable_rewrite_meter = disable_rewrite_meter
        self.boundary_depth = boundary_depth
        self.maximum_dot_count = maximum_dot_count
        self.rewrite_tuplets = rewrite_tuplets
        self.prettify_rewrite_meter = prettify_rewrite_meter
        self.extract_trivial_tuplets = extract_trivial_tuplets
        self.fuse_across_groups_of_beats = fuse_across_groups_of_beats
        self.fuse_quadruple_meter = fuse_quadruple_meter
        self.fuse_triple_meter = fuse_triple_meter
        self.after_rest = after_rest
        self.after_rest_in_new_measure = after_rest_in_new_measure
        self.use_multimeasure_rests = use_multimeasure_rests
        super().__init__(
            head_position=head_position,
            window_size=window_size,
            step_size=step_size,
            max_steps=max_steps,
            repetition_chance=repetition_chance,
            forward_bias=forward_bias,
            process_on_first_call=process_on_first_call,
        )

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return abjad.lilypond(self._contents)

    def __len__(self) -> int:
        r"""Returns the length of :attr:`contents` in terms of
        :attr:`step_size`.
        """
        return ceil(self._contents_length / self._step_size)

    def __call__(self) -> abjad.Selection:
        r"""Calls the looping process for one iteration, returning an
        |abjad.Selection|.
        """
        return super().__call__()

    def __next__(self) -> abjad.Selection:
        r"""Calls the looping process for one iteration, returning an
        |abjad.Selection|.
        """
        return super().__next__()

    ### PRIVATE METHODS ###

    def _slice_contents(self) -> None:
        r"""This method takes a slice of size :attr:`window_size` out of
        :attr:`contents` starting at the current :attr:`head_position`.
        """
        head = self._head_position
        window_size = self._window_size
        dummy_container = abjad.mutate.copy(self._contents_no_time_signature)
        # splitting leaves at both slicing points
        if head > abjad.Duration(0):
            abjad.mutate.split(
                dummy_container[:],
                [head, window_size.duration],
            )
        else:
            abjad.mutate.split(dummy_container[:], [window_size.duration])
        # finding start and end indeces for the window
        for start in range(len(dummy_container)):
            if abjad.get.duration(dummy_container[: start + 1]) > head:
                break
        for end in range(start + 1, len(dummy_container)):
            if abjad.get.duration(dummy_container[start:end]) == window_size.duration:
                break
        else:
            end = len(dummy_container)
        self._notate_music(dummy_container, start, end)

    def _notate_music(
        self,
        dummy_container: abjad.Container,
        start: int,
        end: int,
    ) -> None:
        r"""Handles the notation aspects of the looping window."""
        window_size = self._window_size
        # passing on indicators from the head of an initial splitted leaf
        for index in range(start - 1, -1, -1):
            if abjad.get.indicator(dummy_container[index], abjad.Tie):
                if index == 0 or not abjad.get.indicator(
                    dummy_container[index - 1],
                    abjad.Tie,
                ):
                    for indicator in abjad.get.indicators(
                        dummy_container[index],
                    ):
                        if (
                            not isinstance(indicator, (abjad.TimeSignature, abjad.Tie))
                            and abjad.get.indicator(
                                dummy_container[start],
                                type(indicator),
                            )
                            is None
                        ):
                            abjad.attach(indicator, dummy_container[start])
        # removing ties generated by the split mutation
        abjad.detach(abjad.Tie, dummy_container[start - 1])
        abjad.detach(abjad.Tie, dummy_container[end - 1])
        # handling initial dynamics and slurs
        start_head = abjad.select(dummy_container[start:]).logical_tie(0)[0]
        start_tail = abjad.select(dummy_container[start:]).logical_tie(0)[-1]
        if (
            abjad.get.indicator(start_head, abjad.StartSlur) is None
            and abjad.get.indicator(start_tail, abjad.StopSlur) is None
        ):
            for leaf in dummy_container[start - 1 :: -1].leaves():
                if abjad.get.indicator(leaf, abjad.StartSlur) is not None:
                    abjad.attach(abjad.StartSlur(), start_head)
                    break
                elif abjad.get.indicator(leaf, abjad.StopSlur) is not None:
                    break
        if abjad.get.indicator(start_head, abjad.Dynamic) is None and not isinstance(
            start_head,
            (
                abjad.Rest,
                abjad.MultimeasureRest,
            ),
        ):
            for leaf in dummy_container[start - 1 :: -1].leaves():
                dynamic = abjad.get.indicator(leaf, abjad.Dynamic)
                if dynamic is not None:
                    abjad.attach(dynamic, start_head)
                    break
        # appending rests if necessary
        contents_dur = abjad.get.duration(dummy_container[start:end])
        if contents_dur < window_size.duration:
            missing_dur = window_size.duration - contents_dur
            rests = abjad.LeafMaker()(None, missing_dur)
            dummy_container.extend(rests)
            end += len(rests)
        # transforming abjad.Selection -> abjad.Container for rewrite_meter
        dummy_container = abjad.Container(abjad.mutate.copy(dummy_container[start:end]))
        # handling after rests and time signatures
        time_signature_duration = window_size.duration
        if self._after_rest > 0:
            if (
                self._after_rest_in_new_measure
                and self._use_multimeasure_rests
                and not self._omit_time_signatures
            ):
                if self._after_rest == 1:
                    multiplier = None
                else:
                    multiplier = abjad.Multiplier(self._after_rest)
                rest = abjad.MultimeasureRest(
                    (4, 4),
                    multiplier=multiplier,
                )
            else:
                rest = abjad.LeafMaker()([None], [self._after_rest])
            if not self._after_rest_in_new_measure:
                time_signature_duration += self._after_rest
            else:
                time_signature = abjad.TimeSignature(self._after_rest)
                time_signature.simplify_ratio()
                abjad.attach(time_signature, abjad.select(rest).leaf(0))
            dummy_container.append(rest)
        time_signature = abjad.TimeSignature(time_signature_duration)
        time_signature.simplify_ratio()
        abjad.attach(time_signature, abjad.select(dummy_container).leaf(0))
        # rewriting meter
        if not self._disable_rewrite_meter:
            mutate.auto_rewrite_meter(
                dummy_container,
                boundary_depth=self._boundary_depth,
                maximum_dot_count=self._maximum_dot_count,
                rewrite_tuplets=self._rewrite_tuplets,
                prettify_rewrite_meter=self._prettify_rewrite_meter,
                extract_trivial_tuplets=self._extract_trivial_tuplets,
                fuse_across_groups_of_beats=self._fuse_across_groups_of_beats,
                fuse_quadruple_meter=self._fuse_quadruple_meter,
                fuse_triple_meter=self._fuse_triple_meter,
            )
        mutate.reposition_dynamics(dummy_container[:])
        mutate.reposition_slurs(dummy_container[:])
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    def _get_lilypond_format(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return self.__repr__()

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> abjad.Container:
        r"""The |abjad.Container| to be sliced and looped."""
        return abjad.mutate.copy(self._contents)

    @contents.setter
    def contents(
        self,
        contents: abjad.Container,
    ) -> None:
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or child class")
        if not abjad.select(contents).leaves().are_contiguous_logical_voice():
            raise ValueError("'contents' must be contiguous logical voice")
        if isinstance(contents, abjad.Score):
            self._contents = abjad.mutate.copy(contents[0])
        elif isinstance(contents, abjad.Tuplet):
            self._contents = abjad.Container([abjad.mutate.copy(contents)])
        else:
            self._contents = abjad.mutate.copy(contents)
        self._contents_length = abjad.get.duration(self._contents[:])
        self._contents_no_time_signature = abjad.mutate.copy(self._contents)
        self._remove_all_time_signatures(self._contents_no_time_signature)
        self._is_first_window = True

    @property
    def head_position(self) -> abjad.Duration:
        r"""The position of the head at the start of a looping window."""
        return self._head_position

    @head_position.setter
    def head_position(
        self,
        head_position: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
        ],
    ) -> None:
        r"""This setter method replaces the parent's one since the parent's
        method uses :obj:`int` as input intead of number, :obj:`tuple`, or
        |abjad.Duration|.
        """
        if not isinstance(
            head_position,
            (abjad.Duration, str, tuple, int, float),
        ):
            raise TypeError("'head_position' must be 'abjad.Duration', 'str', 'tuple', or a number")
        if abjad.Duration(head_position) >= self._contents_length:
            raise ValueError("'head_position' must be smaller than the length of 'contents'")
        self._is_first_window = True
        self._head_position = abjad.Duration(head_position)

    @property
    def window_size(self) -> abjad.Meter:
        r"""The length of the looping window."""
        return self._window_size

    @window_size.setter
    def window_size(
        self,
        window_size: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
            abjad.Meter,
        ],
    ) -> None:
        r"""This setter method replaces the parent's one since the parent's
        method uses :obj:`int` as input intead of number, :obj:`tuple`, or
        |abjad.Meter|.
        """
        if not isinstance(
            window_size,
            (
                int,
                float,
                tuple,
                abjad.Duration,
                abjad.Meter,
            ),
        ):
            raise TypeError(
                "'window_size' must be a number, 'tuple', 'abjad.Duration', or 'abjad.Meter'"
            )
        if abjad.Meter(window_size).duration > self._contents_length - self._head_position:
            raise ValueError(
                "'window_size' must be smaller than or equal to the length of 'contents'"
            )
        self._window_size = abjad.Meter(window_size)

    @property
    def step_size(self) -> abjad.Duration:
        r"""The size of each step when moving the head."""
        return self._step_size

    @step_size.setter
    def step_size(
        self,
        step_size: Union[
            int,
            float,
            tuple[int],
            abjad.Duration,
        ],
    ) -> None:
        r"""This setter method replaces the parent's one since the parent's
        method uses :obj:`int` as input intead of number, :obj:`tuple`, or
        |abjad.Duration|.
        """
        if not isinstance(step_size, (abjad.Duration, str, tuple, int, float)):
            raise TypeError("'step_size' must be 'abjad.Duration', 'str', 'tuple', or a number")
        self._step_size = abjad.Duration(step_size)

    @property
    def omit_time_signatures(self) -> bool:
        r"""When ``True``, the output will contain no time signatures."""
        return self._omit_time_signatures

    @omit_time_signatures.setter
    def omit_time_signatures(
        self,
        omit_time_signatures: bool,
    ) -> None:
        if not isinstance(omit_time_signatures, bool):
            raise TypeError("'omit_time_signatures' must be 'bool'")
        self._omit_time_signatures = omit_time_signatures

    @property
    def disable_rewrite_meter(self) -> bool:
        r"""When ``True``, the durations of the notes in the output will not be
        rewritten by the |abjad.Meter.rewrite_meter()| mutation. Rests will
        have the same duration as the logical ties they replaced.
        """
        return self._disable_rewrite_meter

    @disable_rewrite_meter.setter
    def disable_rewrite_meter(
        self,
        disable_rewrite_meter: bool,
    ) -> None:
        if not isinstance(disable_rewrite_meter, bool):
            raise TypeError("'disable_rewrite_meter' must be 'bool'")
        self._disable_rewrite_meter = disable_rewrite_meter

    @property
    def fill_with_rests(self) -> bool:
        r"""When ``True``, the output will contain no time signatures."""
        return self._fill_with_rests

    @fill_with_rests.setter
    def fill_with_rests(
        self,
        fill_with_rests: bool,
    ) -> None:
        if not isinstance(fill_with_rests, bool):
            raise TypeError("'fill_with_rests' must be 'bool'")
        self._fill_with_rests = fill_with_rests

    @property
    def boundary_depth(self) -> Union[int, None]:
        r"""Sets the argument ``boundary_depth`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._boundary_depth

    @boundary_depth.setter
    def boundary_depth(
        self,
        boundary_depth: Optional[int],
    ) -> None:
        if boundary_depth is not None:
            if not isinstance(boundary_depth, int):
                raise TypeError("'boundary_depth' must be 'int'")
        self._boundary_depth = boundary_depth

    @property
    def maximum_dot_count(self) -> Union[int, None]:
        r"""Sets the argument ``maximum_dot_count`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._maximum_dot_count

    @maximum_dot_count.setter
    def maximum_dot_count(
        self,
        maximum_dot_count: Optional[int],
    ) -> None:
        if maximum_dot_count is not None:
            if not isinstance(maximum_dot_count, int):
                raise TypeError("'maximum_dot_count' must be 'int'")
        self._maximum_dot_count = maximum_dot_count

    @property
    def rewrite_tuplets(self) -> bool:
        r"""Sets the argument ``rewrite_tuplets`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._rewrite_tuplets

    @rewrite_tuplets.setter
    def rewrite_tuplets(
        self,
        rewrite_tuplets: bool,
    ) -> None:
        if not isinstance(rewrite_tuplets, bool):
            raise TypeError("'rewrite_tuplets' must be 'bool'")
        self._rewrite_tuplets = rewrite_tuplets

    @property
    def prettify_rewrite_meter(self) -> bool:
        r"""Used to enable or disable the mutation
        |auxjad.mutate.prettify_rewrite_meter()| (default ``True``).
        """
        return self._prettify_rewrite_meter

    @prettify_rewrite_meter.setter
    def prettify_rewrite_meter(
        self,
        prettify_rewrite_meter: bool,
    ) -> None:
        if not isinstance(prettify_rewrite_meter, bool):
            raise TypeError("'prettify_rewrite_meter' must be 'bool'")
        self._prettify_rewrite_meter = prettify_rewrite_meter

    @property
    def extract_trivial_tuplets(self) -> bool:
        r"""Sets the argument ``extract_trivial_tuplets`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._extract_trivial_tuplets

    @extract_trivial_tuplets.setter
    def extract_trivial_tuplets(
        self,
        extract_trivial_tuplets: bool,
    ) -> None:
        if not isinstance(extract_trivial_tuplets, bool):
            raise TypeError("'extract_trivial_tuplets' must be 'bool'")
        self._extract_trivial_tuplets = extract_trivial_tuplets

    @property
    def fuse_across_groups_of_beats(self) -> bool:
        r"""Sets the argument ``fuse_across_groups_of_beats`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_across_groups_of_beats

    @fuse_across_groups_of_beats.setter
    def fuse_across_groups_of_beats(
        self,
        fuse_across_groups_of_beats: bool,
    ) -> None:
        if not isinstance(fuse_across_groups_of_beats, bool):
            raise TypeError("'fuse_across_groups_of_beats' must be 'bool'")
        self._fuse_across_groups_of_beats = fuse_across_groups_of_beats

    @property
    def fuse_quadruple_meter(self) -> bool:
        r"""Sets the argument ``fuse_quadruple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_quadruple_meter

    @fuse_quadruple_meter.setter
    def fuse_quadruple_meter(
        self,
        fuse_quadruple_meter: bool,
    ) -> None:
        if not isinstance(fuse_quadruple_meter, bool):
            raise TypeError("'fuse_quadruple_meter' must be 'bool'")
        self._fuse_quadruple_meter = fuse_quadruple_meter

    @property
    def fuse_triple_meter(self) -> bool:
        r"""Sets the argument ``fuse_triple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_triple_meter

    @fuse_triple_meter.setter
    def fuse_triple_meter(
        self,
        fuse_triple_meter: bool,
    ) -> None:
        if not isinstance(fuse_triple_meter, bool):
            raise TypeError("'fuse_triple_meter' must be 'bool'")
        self._fuse_triple_meter = fuse_triple_meter

    @property
    def after_rest(self) -> abjad.Duration:
        r"""Sets the length of the rest appended at the end of the window
        (default is ``0``).
        """
        return self._after_rest

    @after_rest.setter
    def after_rest(
        self,
        after_rest: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
            abjad.Rest,
        ],
    ) -> None:
        if not isinstance(
            after_rest,
            (abjad.Duration, abjad.Rest, str, tuple, int, float),
        ):
            raise TypeError(
                "'after_rest' must be 'abjad.Duration', "
                "'abjad.Rest', 'str', 'tuple', or a number"
            )
        if isinstance(after_rest, abjad.Rest):
            after_rest = abjad.get.duration(after_rest)
        self._after_rest = abjad.Duration(after_rest)

    @property
    def after_rest_in_new_measure(self) -> bool:
        r"""If ``True``, then after rests will be added to their own measure
        (default is ``False``).
        """
        return self._after_rest_in_new_measure

    @after_rest_in_new_measure.setter
    def after_rest_in_new_measure(
        self,
        after_rest_in_new_measure: bool,
    ) -> None:
        if not isinstance(after_rest_in_new_measure, bool):
            raise TypeError("'after_rest_in_new_measure' must be 'bool'")
        self._after_rest_in_new_measure = after_rest_in_new_measure

    @property
    def use_multimeasure_rests(self) -> bool:
        r"""If ``True``, then multi-measure rests will be used for after rests
        when added to their own measure (default is ``True``).
        """
        return self._use_multimeasure_rests

    @use_multimeasure_rests.setter
    def use_multimeasure_rests(
        self,
        use_multimeasure_rests: bool,
    ) -> None:
        if not isinstance(use_multimeasure_rests, bool):
            raise TypeError("'use_multimeasure_rests' must be 'bool'")
        self._use_multimeasure_rests = use_multimeasure_rests

    ### PRIVATE PROPERTIES ###

    @property
    def _done(self) -> bool:
        r""":obj:`bool` indicating whether the process is done (i.e. whether
        the head position has overtaken the :attr:`contents`'s length).

        This property replaces the parent's one since the parent's property
        uses the number of indeces of :attr:`contents`.
        """
        if self._fill_with_rests:
            return self._head_position >= self._contents_length or self._head_position < 0
        else:
            return (
                self._head_position > self._contents_length - self._window_size.duration
                or self._head_position < 0
            )



import abjad

from .. import get, mutate


def contract_notes(
    container: abjad.Container,
    max_contraction_duration: abjad.Duration,
    *,
    minimum_duration: abjad.Duration = abjad.Duration((0, 1)),
    use_multimeasure_rests: bool = True,
    rewrite_meter: bool = True,
) -> None:
    r"""Mutates an input |abjad.Container| (or child class) in place and has no
    return value; this function contracts all logical ties (notes and chords)
    by a maximum contraction duration.

    Basic usage:
        This function will contract the duration of each logical tie up by the
        allowed maximum contraction duration.

        >>> staff = abjad.Staff(r"c'4 r2. d'2 r2 e'2. r4 f'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                r2.
                d'2
                r2
                e'2.
                r4
                f'1
            }


        ..  figure:: ../_images/contract_notes-DtvjEcpLMF.png

        >>> auxjad.mutate.contract_notes(staff, abjad.Duration((1, 8)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'8
                r2..
                d'4.
                r8
                r2
                e'2
                ~
                e'8
                r4.
                f'2..
                r8
            }

        ..  figure:: ../_images/contract_notes-PEgTVggwUR.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.contract_notes(staff)
        >>> abjad.mutate.contract_notes(staff)

    Chords:
        This function works with chords:

        >>> staff = abjad.Staff(r"\time 3/4 c'2. <d' e' f'>2 r4 g'4 r2")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                <d' e' f'>2
                r4
                g'4
                r2
            }

        ..  figure:: ../_images/contract_notes-MtQM9vn5rV.png

        >>> auxjad.mutate.contract_notes(staff, abjad.Duration((1, 16)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2
                ~
                c'8.
                r16
                <d' e' f'>4..
                r16
                r4
                g'8.
                r16
                r2
            }

        ..  figure:: ../_images/contract_notes-3KViO6UZ74.png

    Second positional argument:
        The second positional argument must be an |abjad.Duration| or objects
        that can instantiate an |abjad.Duration|, such as :obj:`str`,
        :obj:`tuple`, :obj:`int`, and :obj:`float`. The notes will be contract
        by up to that duration, or until the note is removed:

        >>> staff = abjad.Staff(
        ...     r"c'4 r2. d'2 r2 e'2. r4 f'1"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                r2.
                d'2
                r2
                e'2.
                r4
                f'1
            }

        ..  figure:: ../_images/contract_notes-OgFmg6p3VP.png

        >>> auxjad.mutate.contract_notes(staff, abjad.Duration((1, 2)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                R1
                R1
                e'4
                r2.
                f'2
                r2
            }

        ..  figure:: ../_images/contract_notes-YfZ3maHywo.png

    Dynamics:
        Dynamics are preserved:

        >>> staff = abjad.Staff(r"c'4\ppp r2. d'2\ff r2 e'2.\f r4 f'1\mp")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \ppp
                r2.
                d'2
                \ff
                r2
                e'2.
                \f
                r4
                f'1
                \mp
            }

        ..  figure:: ../_images/contract_notes-zKIFJdZuSX.png

        >>> auxjad.mutate.contract_notes(staff, abjad.Duration((1, 8)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'8
                \ppp
                r2..
                d'4.
                \ff
                r8
                r2
                e'2
                \f
                ~
                e'8
                r4.
                f'2..
                \mp
                r8
            }

        ..  figure:: ../_images/contract_notes-iKVB9DmxJf.png

    Time signatures changes:
        This function handles time signature changes:

        >>> staff = abjad.Staff(
        ...     r"\time 3/4 c'4 r2 "
        ...     r"\time 2/4 d'2 "
        ...     r"\time 3/4 e'2. "
        ...     r"\time 4/4 f'1"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                r2
                \time 2/4
                d'2
                \time 3/4
                e'2.
                \time 4/4
                f'1
            }

        ..  figure:: ../_images/contract_notes-cpRcyhRgB7.png

        >>> auxjad.mutate.contract_notes(staff, abjad.Duration((3, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'8
                r8
                r2
                \time 2/4
                d'4.
                r8
                \time 3/4
                e'2
                ~
                e'8
                r8
                \time 4/4
                f'2..
                r8
            }

        ..  figure:: ../_images/contract_notes-R6hnRYZU68.png

    ``minimum_duration``:
        Use the argument ``minimum_duration`` to ensure that notes will not be
        contracted further than this value:

        >>> staff = abjad.Staff(r"c'4 r2. d'2 r2 e'2. r4 f'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                r2.
                d'2
                r2
                e'2.
                r4
                f'1
            }

        ..  figure:: ../_images/contract_notes-lzHgoaJ2Es.png

        >>> auxjad.mutate.contract_notes(staff,
        ...                            abjad.Duration((1, 2)),
        ...                            minimum_duration=abjad.Duration((1, 8)),
        ...                            )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'8
                r2..
                d'8
                r2..
                e'4
                r2.
                f'2
                r2
            }

        ..  figure:: ../_images/contract_notes-o5lELCkJm8.png

    ``use_multimeasure_rests``:
        By default, this function uses multi-measure rests

        >>> staff = abjad.Staff(
        ...     r"\time 4/4 c'4 r2. \time 3/4 d'4 r2 \time 4/4 e'4 r2."
        ... )
        >>> auxjad.mutate.contract_notes(staff, abjad.Duration((1, 2)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                R1
                \time 3/4
                R1 * 3/4
                \time 4/4
                R1
            }

        ..  figure:: ../_images/contract_notes-olfdBpIydF.png


        Set the keyword argument ``use_multimeasure_rests`` to ``False`` to
        disable this behaviour.

        >>> staff = abjad.Staff(
        ...     r"\time 4/4 c'4 r2. \time 3/4 d'4 r2 \time 4/4 e'4 r2."
        ... )
        >>> auxjad.mutate.contract_notes(staff,
        ...                              abjad.Duration((1, 2)),
        ...                              use_multimeasure_rests=False,
        ...                              )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                r1
                \time 3/4
                r2.
                \time 4/4
                r1
            }

        ..  figure:: ../_images/contract_notes-VNPRYNnFl5.png

    ``rewrite_meter``
        By default, this function applies |auxjad.mutate.auto_rewrite_meter()|
        at the end of its process.

        >>> staff = abjad.Staff(r"\time 3/4 r8 c'8 ~ c'2 r8 d'8 ~ d'2")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                r8
                c'8
                ~
                c'2
                r8
                d'8
                ~
                d'2
            }

        ..  figure:: ../_images/contract_notes-rGCxpXl3rI.png

        >>> auxjad.mutate.contract_notes(staff, abjad.Duration((1, 8)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                r8
                c'8
                ~
                c'4.
                r8
                r8
                d'8
                ~
                d'4.
                r8
            }

        ..  figure:: ../_images/contract_notes-NoBsRNQJos.png

        Set ``rewrite_meter`` to ``False`` to disable this. The main reason for
        doing this is performance: when multiple mutations are being applied,
        it is faster to rewrite the meter just once at the end.

        >>> staff = abjad.Staff(r"\time 3/4 r8 c'8 ~ c'2 r8 d'8 ~ d'2")
        >>> auxjad.mutate.contract_notes(staff,
        ...                            abjad.Duration((1, 8)),
        ...                            rewrite_meter=False,
        ...                            )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                r8
                c'2
                r8
                r8
                d'2
                r8
            }

        ..  figure:: ../_images/contract_notes-VNHowP6bFi.png

    ..  warning::

        This function does not support tuplets. Using a container with one or
        more tuplets will result in a :exc:`ValueError` exception being raised:

        >>> staff = abjad.Staff(r"c'4 r4 \times 2/3 {r4 d'4 r4} e'4 r2.")
        >>> auxjad.mutate.contract_notes(staff, abjad.Duration((1, 8)))
        ValueError: first positional argument contains one ore more tuplets,
        which are not currently supported
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("first positional argument must be 'abjad.Container' or child class")
    if len(abjad.select(container).tuplets()) > 0:
        raise ValueError(
            "first positional argument contains one ore more "
            "tuplets, which are not currently supported"
        )
    if not isinstance(
        max_contraction_duration,
        (abjad.Duration, str, tuple, int, float),
    ):
        raise TypeError(
            "second positional argument must be 'abjad.Duration', 'str', 'tuple', or a number"
        )
    if not isinstance(max_contraction_duration, abjad.Duration):
        max_contraction_duration = abjad.Duration(max_contraction_duration)
    if minimum_duration is not None:
        if not isinstance(
            minimum_duration,
            (
                abjad.Duration,
                str,
                tuple,
                int,
                float,
            ),
        ):
            raise TypeError(
                "'minimum_duration' must be 'abjad.Duration', 'str', 'tuple', or a number"
            )
        if not isinstance(minimum_duration, abjad.Duration):
            minimum_duration = abjad.Duration(minimum_duration)
    if not isinstance(use_multimeasure_rests, bool):
        raise TypeError("'use_multimeasure_rests' must be 'bool'")
    if not isinstance(rewrite_meter, bool):
        raise TypeError("'rewrite_meter' must be 'bool'")

    time_signatures = get.time_signature_list(
        container,
        implicit_common_time=False,
    )
    for logical_tie in abjad.select(container).logical_ties():
        if isinstance(logical_tie.head, (abjad.Rest, abjad.MultimeasureRest)):
            continue
        total_duration = abjad.get.duration(logical_tie)
        if total_duration <= minimum_duration:
            continue
        rest_duration = min(max_contraction_duration, total_duration - minimum_duration)
        if rest_duration == abjad.Duration((0, 1)):
            continue
        pitched_duration = total_duration - rest_duration

        if pitched_duration > abjad.Duration((0, 1)):
            if isinstance(logical_tie.head, abjad.Note):
                pitches = logical_tie.head.written_pitch
            elif isinstance(logical_tie.head, abjad.Chord):
                pitches = [pitch for pitch in logical_tie.head.written_pitches]
            replacement_items = abjad.LeafMaker()(
                [pitches, None],
                [pitched_duration, rest_duration],
            )
            indicators = abjad.get.indicators(logical_tie.head)
            if indicators is not None:
                for indicator in indicators:
                    if isinstance(indicator, abjad.Tie):
                        continue
                    abjad.attach(indicator, replacement_items[0])
        else:
            replacement_items = abjad.LeafMaker()([None], [rest_duration])
        abjad.mutate.replace(logical_tie, replacement_items)
    mutate.enforce_time_signature(
        container,
        time_signatures,
        disable_rewrite_meter=True,
    )
    if rewrite_meter:
        mutate.auto_rewrite_meter(container)
    if use_multimeasure_rests:
        mutate.rests_to_multimeasure_rest(container[:])



import abjad

from .. import get, mutate


def _groupped_logical_ties(container) -> list:
    r"""Selects logical ties with all consecutive rests."""
    groupped_logical_ties = []
    selection = abjad.Selection()
    for logical_tie in abjad.select(container).logical_ties():
        if not isinstance(
            logical_tie.head,
            (
                abjad.Rest,
                abjad.MultimeasureRest,
            ),
        ):
            if len(selection) > 0:
                groupped_logical_ties.append(selection)
            selection = abjad.Selection(logical_tie)
        else:
            selection += abjad.Selection(logical_tie)
    if len(selection) > 0:
        groupped_logical_ties.append(selection)
    return groupped_logical_ties


def extend_notes(
    container: abjad.Container,
    max_note_duration: abjad.Duration,
    *,
    gap: abjad.Duration = abjad.Duration((0, 1)),
    gap_before_end: bool = False,
    use_multimeasure_rests: bool = True,
    rewrite_meter: bool = True,
) -> None:
    r"""Mutates an input |abjad.Container| (or child class) in place and has no
    return value; this function extends all logical ties (notes and chords) up
    to a given maximum note duration.

    Basic usage:
        This function will extend the duration of each logical tie up until the
        allowed maximum duration.

        >>> staff = abjad.Staff(r"c'16 r2... d'8 r2.. e'8. r16 r2. f'4 r2.")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                r2...
                d'8
                r2..
                e'8.
                r16
                r2.
                f'4
                r2.
            }

        ..  figure:: ../_images/extend_notes-gJGtjtm3fu.png

        >>> auxjad.mutate.extend_notes(staff, abjad.Duration((1, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                r2.
                d'4
                r2.
                e'4
                r2.
                f'4
                r2.
            }

        ..  figure:: ../_images/extend_notes-6pjHLREOap.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.extend_notes(staff)
        >>> abjad.mutate.extend_notes(staff)

    Longer notes:
        This function will not alter the length of logical ties that are
        already equal to or larger than the maximum extension duration.

        >>> staff = abjad.Staff(r"c'16 r2... d'2 r2 e'2. r4 f'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                r2...
                d'2
                r2
                e'2.
                r4
                f'1
            }

        ..  figure:: ../_images/extend_notes-56rOAsxHPP.png

        >>> auxjad.mutate.extend_notes(staff, abjad.Duration((1, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                r2.
                d'2
                r2
                e'2.
                r4
                f'1
            }

        ..  figure:: ../_images/extend_notes-egF496w5Bs.png

    Chords:
        This function works with chords:

        >>> staff = abjad.Staff(
        ...     r"\time 3/4 c'8 r8 r2 r4 <d' e' f'>4 r4 r8 g'16 r16 r2"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'8
                r8
                r2
                r4
                <d' e' f'>4
                r4
                r8
                g'16
                r16
                r2
            }

        ..  figure:: ../_images/extend_notes-2151PE127t.png

        >>> auxjad.mutate.extend_notes(staff, abjad.Duration((2, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2
                r4
                r4
                <d' e' f'>2
                r8
                g'8
                ~
                g'4.
                r8
            }

        ..  figure:: ../_images/extend_notes-GVM7SBRQ5z.png

    Second positional argument:
        The second positional argument must be an |abjad.Duration| or objects
        that can instantiate an |abjad.Duration|, such as :obj:`str`,
        :obj:`tuple`, :obj:`int`, and :obj:`float`. The notes will be extended
        up to that duration, or until the next note if the full duration is not
        possible:

        >>> staff = abjad.Staff(
        ...     r"c'16 r4.. d'16 r4.. e'16 r2... f'16 r4.. g'16 r4.."
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                r4..
                d'16
                r4..
                e'16
                r2...
                f'16
                r4..
                g'16
                r4..
            }

        ..  figure:: ../_images/extend_notes-GfrvgpqH2J.png

        >>> auxjad.mutate.extend_notes(staff, abjad.Duration((3, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'2
                d'2
                e'2.
                r4
                f'2
                g'2
            }

        ..  figure:: ../_images/extend_notes-1vKtKyK7up.png

    Dynamics:
        Dynamics are preserved:

        >>> staff = abjad.Staff(
        ...     r"c'16\ppp r2... d'16\ff r2... e'16\f r2... f'16\mp r2..."
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                \ppp
                r2...
                d'16
                \ff
                r2...
                e'16
                \f
                r2...
                f'16
                \mp
                r2...
            }

        ..  figure:: ../_images/extend_notes-rAHtip59b2.png

        >>> auxjad.mutate.extend_notes(staff, abjad.Duration((1, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'2.
                \ppp
                r4
                d'2.
                \ff
                r4
                e'2.
                \f
                r4
                f'2.
                \mp
                r4
            }

        ..  figure:: ../_images/extend_notes-TNbei46PhS.png

    Time signatures changes:
        This function handles time signature changes:

        >>> staff = abjad.Staff(
        ...     r"\time 3/4 c'16 r8. r2 "
        ...     r"\time 2/4 d'8 r8 e'8 r8 "
        ...     r"\time 3/4 r2 f'16 r8."
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'16
                r8.
                r2
                \time 2/4
                d'8
                r8
                e'8
                r8
                \time 3/4
                r2
                f'16
                r8.
            }

        ..  figure:: ../_images/extend_notes-k8d4mKMaS7.png

        >>> auxjad.mutate.extend_notes(staff, abjad.Duration((3, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                \time 2/4
                d'4
                e'4
                ~
                \time 3/4
                e'2
                f'4
            }

        ..  figure:: ../_images/extend_notes-5eDO8ehSld.png

    ``gap``:
        Use the argument ``gap`` to ensure that there a minimum gap is left
        between leaves when extending them:

        >>> staff = abjad.Staff(r"c'4 r4 d'4 r4 e'4 r2.")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                r4
                d'4
                r4
                e'4
                r2.
            }

        ..  figure:: ../_images/extend_notes-zpAx8NYXof.png

        >>> auxjad.mutate.extend_notes(staff,
        ...                            abjad.Duration((2, 4)),
        ...                            gap=abjad.Duration((1, 16)),
        ...                            )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4..
                r16
                d'4..
                r16
                e'2
                r2
            }

        ..  figure:: ../_images/extend_notes-t1HIPtyJkn.png

        Note that this function will not shorten leaves, so previous leaves
        that had no gaps will still have no gaps between them. ``gap`` will
        only affect leaves that are extended by the function.

        >>> staff = abjad.Staff(r"c'2 d'4 r4 e'4 r4 f'2 ~ f'2 r2")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'2
                d'4
                r4
                e'4
                r4
                f'2
                ~
                f'2
                r2
            }

        ..  figure:: ../_images/extend_notes-ZfrusN3f8T.png

        >>> auxjad.mutate.extend_notes(staff,
        ...                            abjad.Duration((2, 4)),
        ...                            gap=abjad.Duration((1, 16)),
        ...                            )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'2
                d'4..
                r16
                e'4..
                r16
                f'2
                ~
                f'2
                r2
            }

        ..  figure:: ../_images/extend_notes-5iteVkhz3l.png

    ``gap_before_end``:
        By default, a when ``gap`` is present it is not applied to the last
        pitched at the end of the container.

        >>> staff = abjad.Staff(r"c'16 r8. d'16 r8. e'16 r8. f'16 r8.")
        >>> auxjad.mutate.extend_notes(staff,
        ...                            abjad.Duration((1, 4)),
        ...                            gap=abjad.Duration((1, 16)),
        ...                            )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'8.
                r16
                d'8.
                r16
                e'8.
                r16
                f'4
            }

        ..  figure:: ../_images/extend_notes-LuCN8pgjLH.png

        Set ``gap_before_end`` to ``True`` to ensure that there is also a
        minimum gap at the end of the container.

        >>> staff = abjad.Staff(r"c'16 r8. d'16 r8. e'16 r8. f'16 r8.")
        >>> auxjad.mutate.extend_notes(staff,
        ...                            abjad.Duration((1, 4)),
        ...                            gap=abjad.Duration((1, 16)),
        ...                            gap_before_end=True,
        ...                            )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'8.
                r16
                d'8.
                r16
                e'8.
                r16
                f'8.
                r16
            }

        ..  figure:: ../_images/extend_notes-LppuegmkEP.png

    ``use_multimeasure_rests``:
        By default, this function uses multi-measure rests

        >>> staff = abjad.Staff(r"\time 3/4 r8 c'8 r4 c'4 r2. r8 c'8 r2 r2.")
        >>> auxjad.mutate.extend_notes(staff, abjad.Duration((2, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                r8
                c'4.
                c'4
                ~
                c'4
                r2
                r8
                c'8
                ~
                c'4.
                r8
                R1 * 3/4
            }

        ..  figure:: ../_images/extend_notes-WaLYhL24UG.png


        Set the keyword argument ``use_multimeasure_rests`` to ``False`` to
        disable this behaviour.

        >>> staff = abjad.Staff(r"\time 3/4 r8 c'8 r4 c'4 r2. r8 c'8 r2 r2.")
        >>> auxjad.mutate.extend_notes(staff,
        ...                            abjad.Duration((2, 4)),
        ...                            use_multimeasure_rests=False,
        ...                            )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                r8
                c'4.
                c'4
                ~
                c'4
                r2
                r8
                c'8
                ~
                c'4.
                r8
                r2.
            }

        ..  figure:: ../_images/extend_notes-H1uDCqQ8iz.png

    ``rewrite_meter``
        By default, this function applies |auxjad.mutate.auto_rewrite_meter()|
        at the end of its process.

        >>> staff = abjad.Staff(r"c'16 r4.. d'16 r4..")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                r4..
                d'16
                r4..
            }

        ..  figure:: ../_images/extend_notes-BcAYS4SvUL.png

        >>> auxjad.mutate.extend_notes(staff, abjad.Duration((1, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                r4
                d'4
                r4
            }

        ..  figure:: ../_images/extend_notes-QNEks8lUXB.png

        Set ``rewrite_meter`` to ``False`` to disable this. The main reason for
        doing this is performance: when multiple mutations are being applied,
        it is faster to rewrite the meter just once at the end.

        >>> staff = abjad.Staff(r"c'16 r4.. d'16 r4..")
        >>> auxjad.mutate.extend_notes(staff,
        ...                            abjad.Duration((1, 4)),
        ...                            rewrite_meter=False,
        ...                            )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                ~
                c'8.
                r4
                d'16
                ~
                d'8.
                r4
            }

        ..  figure:: ../_images/extend_notes-49jhNoSa9T.png

    ..  warning::

        This function does not support tuplets. Using a container with one or
        more tuplets will result in a :exc:`ValueError` exception being raised:

        >>> staff = abjad.Staff(r"c'4 r4 \times 2/3 {r4 d'4 r4} e'4 r2.")
        >>> auxjad.mutate.extend_notes(staff, abjad.Duration((1, 4)))
        ValueError: first positional argument contains one ore more tuplets,
        which are not currently supported
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("first positional argument must be 'abjad.Container' or child class")
    if len(abjad.select(container).tuplets()) > 0:
        raise ValueError(
            "first positional argument contains one ore more "
            "tuplets, which are not currently supported"
        )
    if not isinstance(
        max_note_duration,
        (abjad.Duration, str, tuple, int, float),
    ):
        raise TypeError(
            "second positional argument must be 'abjad.Duration', 'str', 'tuple', or a number"
        )
    if not isinstance(max_note_duration, abjad.Duration):
        max_note_duration = abjad.Duration(max_note_duration)
    if gap is not None:
        if not isinstance(gap, (abjad.Duration, str, tuple, int, float)):
            raise TypeError("'gap' must be 'abjad.Duration', 'str', 'tuple', or a number")
        if not isinstance(gap, abjad.Duration):
            gap = abjad.Duration(gap)
    if not isinstance(gap_before_end, bool):
        raise TypeError("'gap_before_end' must be 'bool'")
    if not isinstance(use_multimeasure_rests, bool):
        raise TypeError("'use_multimeasure_rests' must be 'bool'")
    if not isinstance(rewrite_meter, bool):
        raise TypeError("'rewrite_meter' must be 'bool'")

    time_signatures = get.time_signature_list(
        container,
        implicit_common_time=False,
    )
    groups = _groupped_logical_ties(container)
    for group_index, group in enumerate(groups):
        if not gap_before_end and group_index == len(groups) - 1:
            gap = abjad.Duration((0, 1))
        if isinstance(group[0], (abjad.Rest, abjad.MultimeasureRest)):
            continue
        pitched_duration = 0
        rest_duration = 0
        for item in group:
            if isinstance(item, (abjad.Note, abjad.Chord)):
                pitched_duration += abjad.get.duration(item)
            else:
                rest_duration += abjad.get.duration(item)
        missing_duration = max_note_duration - pitched_duration
        available_duration = rest_duration - gap
        if missing_duration <= 0 or available_duration <= 0:
            continue
        else:
            extended_duration = min(missing_duration, available_duration)
            new_rest_duration = rest_duration - extended_duration

        for index, item in enumerate(group):
            if isinstance(item, (abjad.Rest, abjad.MultimeasureRest)):
                first_rest_index = index
                break
        if isinstance(group[first_rest_index - 1], abjad.Note):
            pitches = group[first_rest_index - 1].written_pitch
        elif isinstance(group[first_rest_index - 1], abjad.Chord):
            pitches = [pitch for pitch in group[first_rest_index - 1].written_pitches]
        if new_rest_duration > abjad.Duration((0, 1)):
            replacement_items = abjad.LeafMaker()(
                [pitches, None],
                [extended_duration, new_rest_duration],
            )
        else:
            replacement_items = abjad.LeafMaker()(
                [pitches],
                [extended_duration],
            )
        abjad.mutate.replace(group[first_rest_index:], replacement_items)
        abjad.attach(abjad.Tie(), group[first_rest_index - 1])
    mutate.enforce_time_signature(
        container,
        time_signatures,
        disable_rewrite_meter=True,
    )
    if rewrite_meter:
        mutate.auto_rewrite_meter(container)
    if use_multimeasure_rests:
        mutate.rests_to_multimeasure_rest(container[:])



from collections.abc import Iterable
from typing import Union

import abjad


def selections_are_identical(
    selections: Union[
        Iterable[abjad.Component],
        Iterable[abjad.Selection],
    ],
    *,
    include_indicators: bool = True,
) -> bool:
    r"""Returns a :obj:`bool` representing whether two or more selections are
    identical or not. Input argument must be an iterable made of two or more
    |abjad.Selection|'s.

    Basic usage:
        When the pitches and effective durations of all leaves in all
        selections are identical, this function returns ``True``:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.selections_are_identical(selections)
        True

        This function can handle multiple selections, and will compare them
        among each other returning ``True`` if all are identical:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container3 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container4 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> selections = [
        ...     container1[:],
        ...     container2[:],
        ...     container3[:],
        ...     container4[:],
        ... ]
        >>> auxjad.get.selections_are_identical(selections)
        True

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.get|. It can thus be used from either |auxjad.get|_ or
        |abjad.get| namespaces. Therefore, the two lines below are equivalent:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.selections_are_identical(selections)
        True
        >>> abjad.get.selections_are_identical(selections)
        True

    Effective durations:
        Even if all leaves of both selections are identical in relation to both
        pitches and written durations, the function considers the effective
        durations. This means that situations like the one below do not yield a
        false positive:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(
        ...     r"\times 3/2 {c'4 d'4 e'4} f'4 <g' a'>2 r2"
        ... )
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.selections_are_identical(selections)
        False

    ``include_indicators``:
        By default, this function includes indicators in the comparison, so the
        containers in the example below are understood to be different:

        >>> container1 = abjad.Staff(r"c'4\pp d'4 e'4-. f'4 <g' a'>2-> r2")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.selections_are_identical(selections)
        False

        Set the argument ``include_indicators`` to ``False`` to ignore
        indicators when comparison selections. In that case, the containers in
        the example above are then considered identical:

        >>> container1 = abjad.Staff(r"c'4\pp d'4 e'4-. f'4 <g' a'>2-> r2")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.selections_are_identical(
        ...     selections,
        ...     include_indicators=False,
        ... )
        True

    Grace notes:
        This function also handles grace notes.

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Staff(r"c'4 \grace{d'4} d'4 e'4 f'4")
        >>> selection1 = abjad.select(container1)
        >>> selection2 = abjad.select(container2)
        >>> selections = [selection1, selection2]
        >>> auxjad.get.selections_are_identical(selections)
        False

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(
        ...     r"c'4 \grace{c''4} d'4 e'4 f'4 <g' a'>2 r2"
        ... )
        >>> selection1 = abjad.select(container1)
        >>> selection2 = abjad.select(container2)
        >>> selections = [selection1, selection2]
        >>> auxjad.get.selections_are_identical(selections)
        False

        >>> container1 = abjad.Staff(
        ...     r"c'4 \grace{c''4} d'4 e'4 f'4 <g' a'>2 r2"
        ... )
        >>> container2 = abjad.Staff(
        ...     r"c'4 \grace{c''8} d'4 e'4 f'4 <g' a'>2 r2"
        ... )
        >>> selection1 = abjad.select(container1)
        >>> selection2 = abjad.select(container2)
        >>> selections = [selection1, selection2]
        >>> auxjad.get.selections_are_identical(selections)
        False

        >>> container1 = abjad.Staff(
        ...     r"c'4 \grace{c''16} d'4 e'4 f'4 <g' a'>2 r2"
        ... )
        >>> container2 = abjad.Staff(
        ...     r"c'4 \grace{c''16} d'4 e'4 f'4 <g' a'>2 r2"
        ... )
        >>> selection1 = abjad.select(container1)
        >>> selection2 = abjad.select(container2)
        >>> selections = [selection1, selection2]
        >>> auxjad.get.selections_are_identical(selections)
        True

    ..  warning::

        It is important to create selections using |abjad.select()| as shown in
        the example above, instead of using the syntax ``container[:]``, since
        the latter ignores grace notes.

    ..  note::

        It is important to note it is the contents of the containers which are
        compared, so containers of different classes can still return a
        ``True`` value.

        >>> container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> selections = [container1[:], container2[:]]
        >>> auxjad.get.selections_are_identical(selections)
        True
    """
    if not isinstance(selections, Iterable):
        raise TypeError("argument must be an iterable of 'abjad.Selection's or 'abjad.Component's")
    for selection in selections:
        if not isinstance(selection, (abjad.Component, abjad.Selection)):
            raise TypeError(
                "argument must be an iterable of 'abjad.Selection's or 'abjad.Component's"
            )
    if not isinstance(include_indicators, bool):
        raise TypeError("'include_indicators' must be 'bool'")

    for index, selection1 in enumerate(selections[:-1]):
        for selection2 in selections[index + 1 :]:
            leaves1 = [leaf for leaf in selection1.leaves()]
            leaves2 = [leaf for leaf in selection2.leaves()]
            if len(leaves1) != len(leaves2):
                return False
            for leaf1, leaf2 in zip(leaves1, leaves2):
                if not isinstance(leaf1, type(leaf2)):
                    return False
                if abjad.get.duration(leaf1) != abjad.get.duration(leaf2):
                    return False
                if isinstance(leaf1, abjad.Note) and leaf1.written_pitch != leaf2.written_pitch:
                    return False
                if (
                    isinstance(leaf1, abjad.Chord)
                    and leaf1.written_pitches != leaf2.written_pitches
                ):
                    return False
                leaf1_graces = abjad.get.before_grace_container(leaf1)
                leaf2_graces = abjad.get.before_grace_container(leaf2)
                if not isinstance(leaf1_graces, type(leaf2_graces)):
                    return False
                if include_indicators:
                    indicators1 = [format(indicator) for indicator in abjad.get.indicators(leaf1)]
                    indicators2 = [format(indicator) for indicator in abjad.get.indicators(leaf2)]
                    if not all(indicator1 in indicators2 for indicator1 in indicators1):
                        return False
                    if not all(indicator2 in indicators1 for indicator2 in indicators2):
                        return False
    return True



import abjad


def reposition_slurs(
    selection: abjad.Selection,
    *,
    allow_slurs_under_rests: bool = False,
    close_unterminated_final_slur: bool = True,
) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function repositions all slurs that starts or ends on rests.

    Basic usage:
        This function will shift slurs that ends on rests to the previous
        pitched leaf.

        >>> staff = abjad.Staff(r"c'1( d'2 r2) r1 e'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                r2
                )
                r1
                e'1
            }

        ..  figure:: ../_images/reposition_slurs-uxji4xx6ftk.png

        >>> staff = abjad.Staff(r"c'1( d'2 r2) r1 e'1")
        >>> auxjad.mutate.reposition_slurs(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                )
                r2
                r1
                e'1
            }

        ..  figure:: ../_images/reposition_slurs-7nnp5cttm4y.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.reposition_slurs(staff[:])
        >>> abjad.mutate.reposition_slurs(staff[:])

    Rests:
        Slurs starting on rests are shifted to the next pitched leaf.

        >>> staff = abjad.Staff(r"c'1 r2( d'2 e'1)")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                r2
                (
                d'2
                e'1
                )
            }

        ..  figure:: ../_images/reposition_slurs-2j7hgqd7bt1.png

        >>> staff = abjad.Staff(r"c'1 r2( d'2 e'1)")
        >>> auxjad.mutate.reposition_slurs(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                r2
                d'2
                (
                e'1
                )
            }

        ..  figure:: ../_images/reposition_slurs-i957u1wt30m.png

    Multiple rests:
        This function also works when multiple rests are present.

        >>> staff = abjad.Staff(r"c'1( d'2 r2 r1) e'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                r2
                r1
                )
                e'1
            }

        ..  figure:: ../_images/reposition_slurs-v76u42x7idk.png

        >>> staff = abjad.Staff(r"c'1( d'2 r2 r1) e'1")
        >>> auxjad.mutate.reposition_slurs(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                )
                r2
                r1
                e'1
            }

        ..  figure:: ../_images/reposition_slurs-burs1t0daid.png

    ``allow_slurs_under_rests``:
        By default, a slur crossing a rest is broken into two.

        >>> staff = abjad.Staff(r"c'1( d'2 r2 e'1 f'1)")
        >>> auxjad.mutate.reposition_slurs(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                )
                r2
                e'1
                (
                f'1
                )
            }

        ..  figure:: ../_images/reposition_slurs-8wb7orpt285.png

        Set the optional keyword argument ``allow_slurs_under_rests`` to
        ``True`` to allow slurs under rests.

        >>> staff = abjad.Staff(r"c'1( d'2 r2 e'1 f'1)")
        >>> auxjad.mutate.reposition_slurs(
        ...     staff[:],
        ...     allow_slurs_under_rests=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                r2
                e'1
                f'1
                )
            }

        ..  figure:: ../_images/reposition_slurs-ftb59kz6u8j.png

    ``close_unterminated_final_slur``:
        By default, unterminated slurs at the end of the selection are closed
        when possible or removed when not.

        >>> staff = abjad.Staff(r"c'1( d'2 r2 e'2 f'2) g'1( a'1")
        >>> auxjad.mutate.reposition_slurs(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                )
                r2
                e'2
                (
                f'2
                )
                g'1
                (
                a'1
                )
            }

        ..  figure:: ../_images/reposition_slurs-70dli8e0kqr.png

        Set the optional keyword argument ``close_unterminated_final_slur`` to
        ``False`` to disable this behaviour.

        >>> staff = abjad.Staff(r"c'1( d'2 r2 e'2 f'2) g'1( a'1")
        >>> auxjad.mutate.reposition_slurs(
        ...     staff[:],
        ...     close_unterminated_final_slur=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                )
                r2
                e'2
                (
                f'2
                )
                g'1
                (
                a'1
            }

        ..  figure:: ../_images/reposition_slurs-1usa2dezl45.png

        When there are no pitched leaves left after an unterminated open slur,
        it is removed.

        >>> staff = abjad.Staff(r"c'1( d'2 r2 e'2 f'2) g'1( r1")
        >>> auxjad.mutate.reposition_slurs(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                )
                r2
                e'2
                (
                f'2
                )
                g'1
                r1
            }

        ..  figure:: ../_images/reposition_slurs-a0uakfcltuf.png

    ..  note::

        Duplicate slur starts or stops are removed. Note that the score output
        will not change, as LilyPond also ignores duplicate slurs, but the
        output in the ``.ly`` file will be cleaner.

        >>> staff = abjad.Staff(r"c'1( d'2) e'2) f'2( g'2( a'1)")
        >>> auxjad.mutate.reposition_slurs(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                (
                d'2
                )
                e'2
                f'2
                (
                g'2
                a'1
                )
            }

        ..  figure:: ../_images/reposition_slurs-0ugn322x3tr.png

    ..  warning::

        The input selection must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Container' or child class")
    if not selection.leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")
    if not isinstance(allow_slurs_under_rests, bool):
        raise TypeError("'allow_slurs_under_rests' must be 'bool'")
    if not isinstance(close_unterminated_final_slur, bool):
        raise TypeError("'close_unterminated_final_slur' must be 'bool'")

    leaves = selection.leaves()

    # checking for final unfinished slurs
    if close_unterminated_final_slur:
        for leaf in leaves[::-1]:
            if abjad.get.indicator(leaf, abjad.StartSlur) is not None:
                if leaf is leaves[-1]:
                    abjad.detach(abjad.StartSlur, leaf)
                elif abjad.get.indicator(leaves[-1], abjad.StopSlur) is None:
                    abjad.attach(abjad.StopSlur(), leaves[-1])
            if abjad.get.indicator(leaf, abjad.StopSlur) is not None:
                break

    # checking for duplicate open or close slurs
    start_slur_count = 0
    stop_slur_count = 0
    for leaf in leaves:
        if abjad.get.indicator(leaf, abjad.StartSlur) is not None:
            start_slur_count += 1
            stop_slur_count = 0
            if start_slur_count > 1:
                abjad.detach(abjad.StartSlur, leaf)
        elif abjad.get.indicator(leaf, abjad.StopSlur) is not None:
            stop_slur_count += 1
            start_slur_count = 0
            if stop_slur_count > 1:
                abjad.detach(abjad.StopSlur, leaf)

    # shifting slurs from rests to notes
    shifted_startslur = None
    for leaf in leaves:
        if isinstance(leaf, (abjad.Rest, abjad.MultimeasureRest)):
            if abjad.get.indicator(leaf, abjad.StartSlur) is not None:
                shifted_startslur = abjad.get.indicator(leaf, abjad.StartSlur)
                abjad.detach(abjad.StartSlur, leaf)
        else:
            if abjad.get.indicator(leaf, abjad.StartSlur) is None:
                if shifted_startslur is not None:
                    abjad.attach(shifted_startslur, leaf)
                    shifted_startslur = None
    shifted_stopslur = None
    for leaf in leaves[::-1]:
        if isinstance(leaf, (abjad.Rest, abjad.MultimeasureRest)):
            if abjad.get.indicator(leaf, abjad.StopSlur) is not None:
                shifted_stopslur = abjad.get.indicator(leaf, abjad.StopSlur)
                abjad.detach(abjad.StopSlur, leaf)
        else:
            if abjad.get.indicator(leaf, abjad.StopSlur) is None:
                if shifted_stopslur is not None:
                    abjad.attach(shifted_stopslur, leaf)
                    shifted_stopslur = None

    # splitting slurs under rests
    if not allow_slurs_under_rests:
        active_slur = False
        for index, leaf in enumerate(leaves):
            if abjad.get.indicator(leaf, abjad.StartSlur) is not None:
                active_slur = True
            elif abjad.get.indicator(leaf, abjad.StopSlur) is not None:
                if not active_slur:
                    abjad.detach(abjad.StopSlur, leaf)
                active_slur = False
            if isinstance(leaf, (abjad.Rest, abjad.MultimeasureRest)) and active_slur:
                previous_leaf = abjad.select(leaf).with_previous_leaf()[0]
                if abjad.get.indicator(previous_leaf, abjad.StopSlur) is None:
                    abjad.attach(abjad.StopSlur(), previous_leaf)
                    active_slur = False
                for next_leaf in leaves[index + 1 :]:
                    if not isinstance(
                        next_leaf,
                        (
                            abjad.Rest,
                            abjad.MultimeasureRest,
                        ),
                    ):
                        if abjad.get.indicator(next_leaf, abjad.StartSlur) is None:
                            abjad.attach(abjad.StartSlur(), next_leaf)
                        break
        for leaf in leaves:
            if (
                abjad.get.indicator(leaf, abjad.StartSlur) is not None
                and abjad.get.indicator(leaf, abjad.StopSlur) is not None
            ):
                abjad.detach(abjad.StartSlur, leaf)
                abjad.detach(abjad.StopSlur, leaf)

    # removing slurs spanning a single logical tie
    for logical_tie in selection.logical_ties():
        if (
            abjad.get.indicator(logical_tie[0], abjad.StartSlur) is not None
            and abjad.get.indicator(logical_tie[-1], abjad.StopSlur) is not None
        ):
            abjad.detach(abjad.StartSlur, logical_tie[0])
            abjad.detach(abjad.StopSlur, logical_tie[-1])
